import unittest
from cryoloBM_tools import scale, priors2star, coords2warp, rotatecboxcoords
from cryoloBM import boxmanager_tools
from cryolo import CoordsIO
import os
import numpy as np

class MyTestCase(unittest.TestCase):

    def test_scale_fil_seg_file(self):
        input_path = os.path.join(os.path.dirname(__file__),
                                  "../resources/segmented_fila_data/segmented_filament.box")
        output_path = os.path.join(os.path.dirname(__file__),
                                   "../resources/out/")
        scale.scale(input_path=input_path,
                    output_path=output_path,
                    scale_factor=0.5)

        written_file_path = os.path.join(os.path.dirname(__file__),
                                         "../resources/out/segmented_filament.box")
        input_filaments = CoordsIO.read_eman1_helicon(input_path)
        scaled_filaments = CoordsIO.read_eman1_helicon(written_file_path)

        for fil_index, fil in enumerate(input_filaments):
            scaled_fil = scaled_filaments[fil_index]
            for boxindex, box in enumerate(fil.boxes):
                scaled_box = scaled_fil.boxes[boxindex]

                diff = np.abs(scaled_box.x * 1/0.5 - box.x)
                same = diff <= 0.001
                self.assertTrue(same, "Not the same " + str(diff))

                diff = np.abs(scaled_box.y * 1 / 0.5 - box.y)
                same = diff <= 0.001
                self.assertTrue(same, "Not the same " + str(diff))

    def test_scale_fil_seg_folder(self):
        input_path = os.path.join(os.path.dirname(__file__),
                                  "../resources/segmented_fila_data/")
        output_path = os.path.join(os.path.dirname(__file__),
                                   "../resources/out/")
        scale.scale(input_path=input_path,
                    output_path=output_path,
                    scale_factor=0.5)

        written_file_path = os.path.join(os.path.dirname(__file__),
                                         "../resources/out/segmented_filament.box")
        input_filaments = CoordsIO.read_eman1_helicon(os.path.join(input_path,"segmented_filament.box"))
        scaled_filaments = CoordsIO.read_eman1_helicon(written_file_path)

        for fil_index, fil in enumerate(input_filaments):
            scaled_fil = scaled_filaments[fil_index]
            for boxindex, box in enumerate(fil.boxes):
                scaled_box = scaled_fil.boxes[boxindex]

                diff = np.abs(scaled_box.x * 1/0.5 - box.x)
                same = diff <= 0.001
                self.assertTrue(same, "Not the same " + str(diff))

                diff = np.abs(scaled_box.y * 1 / 0.5 - box.y)
                same = diff <= 0.001
                self.assertTrue(same, "Not the same " + str(diff))

    def test_scale_coords_file(self):

        input_path = os.path.join(os.path.dirname(__file__),
                            "../resources/coords_data/example.coords")
        output_path = os.path.join(os.path.dirname(__file__),
                                  "../resources/out/")
        scale.scale(input_path=input_path,
                    output_path=output_path,
                    scale_factor=0.5)


        written_file_path = os.path.join(os.path.dirname(__file__),
                                   "../resources/out/example.coords")
        input_coords = np.atleast_2d(np.genfromtxt(input_path))
        written_coords = np.atleast_2d(np.genfromtxt(written_file_path))

        scal_fact_mat = 2*written_coords-input_coords

        scal_fact_mat = np.abs(scal_fact_mat)<10**-4
        print(scal_fact_mat)
        all_true = np.all(scal_fact_mat)
        os.remove(written_file_path)
        self.assertEqual(True, all_true)

    def test_priors2star(self):

        input_path_star = os.path.join(os.path.dirname(__file__),
                            "../resources/test_priors2star_data/particles.star")

        input_path_fid1 = os.path.join(os.path.dirname(__file__),
                                       "../resources/test_priors2star_data/tomo1_bin4_fid.coords")

        input_path_fid2 = os.path.join(os.path.dirname(__file__),
                                       "../resources/test_priors2star_data/tomo2_bin4_fid.coords")

        output_path = os.path.join(os.path.dirname(__file__),
                                  "../resources/out/particles_with_prior.star")
        if os.path.exists(output_path):
            os.remove(output_path)
        priors2star.add_prior_to_star(input_path_star, [input_path_fid1,input_path_fid2], output_path)
        written = os.path.exists(output_path)
        if os.path.exists(output_path):
            pass #os.remove(output_path)
        self.assertEqual(True, written)

    def test_priors2star_match(self):
        tomonames = ["/a/b/c/str/abc.mrc", "/a/b/c/str/abc.mrc", "/a/b/c/str/abc.mrc", "/a/b/c/str/def.mrc","/a/b/c/str/def.mrc","/a/b/c/str/def.mrc"]

        fid_names = ["/a/b/c/coords/abc_bin4_fid.coords","/a/b/c/coords/def_bin4_fid.coords"]
        fid_index_list = priors2star.match(tomonames, fid_names)
        self.assertListEqual(fid_index_list[0], [0, 1, 2], "First list is different")
        self.assertListEqual(fid_index_list[1], [3, 4, 5], "Second list is different")

    def test_priors2star_match_with_star(self):
        from pyStarDB import sp_pystardb as star
        in_star = os.path.join(os.path.dirname(__file__),
                                      "../resources/test_priors2star_data/particles.star")
        sfile = star.StarFile(in_star)
        relion_dataframe = sfile['']

        tomonames = relion_dataframe['_rlnMicrographName']

        fid_names = ["/a/b/c/coords/tomo1_bin4_fid.coords","/a/b/c/coords/tomo2_bin4_fid.coords"]
        fid_index_list = priors2star.match(tomonames, fid_names)
        self.assertEqual(len(fid_index_list[0]), 1505, "First list is different")
        self.assertEqual(len(fid_index_list[1]), 1505, "Second list is different")


    def test_coords2warp(self):
        input_path_fid = os.path.join(os.path.dirname(__file__),
                                      "../resources/test_coords2warp_data/")

        output_path = os.path.join(os.path.dirname(__file__),
                                   "../resources/out/")
        if os.path.exists(os.path.join(output_path,"particles_warp.star")):
            os.remove(os.path.join(output_path,"particles_warp.star"))

        pd = coords2warp.convert(
            input_path = input_path_fid,
            pixelsize = 1.75,
            magnification = 100000,
            scale = 1
        )
        self.assertEqual(True,pd != None)

    def boxes_to_array(self, boxes):
        arr = np.zeros(shape=(len(boxes),3))
        for box_i, box in enumerate(boxes):
            arr[box_i,0] = box.x
            arr[box_i,1] = box.y
            arr[box_i,2] = box.z
        return arr

    def flatten_list(self, list_of_lists):
        return [item for sublist in list_of_lists for item in sublist]

    def test_scale_cbox_file(self):

        input_path = os.path.join(os.path.dirname(__file__),
                            "../resources/cbox_data/example3D_filament.cbox")
        output_path = os.path.join(os.path.dirname(__file__),
                                  "../resources/out/")
        scale.scale(input_path=input_path,
                    output_path=output_path,
                    scale_factor=0.5)


        written_file_path = os.path.join(os.path.dirname(__file__),
                                   "../resources/out/example3D_filament.cbox")

        input_fil = CoordsIO.read_cbox_boxfile(input_path)
        input_fil_boxes = [fil.boxes for fil in input_fil]
        input_fil_boxes = self.flatten_list(input_fil_boxes)
        input_coords = self.boxes_to_array(input_fil_boxes)

        written_fil = CoordsIO.read_cbox_boxfile(written_file_path)
        written_fil_boxes = [fil.boxes for fil in written_fil]
        written_fil_boxes = self.flatten_list(written_fil_boxes)
        written_coords = self.boxes_to_array(written_fil_boxes)
        scal_fact_mat = 2*written_coords-input_coords

        scal_fact_mat = np.abs(scal_fact_mat)<10**-4
        all_true = np.all(scal_fact_mat)
        os.remove(written_file_path)
        self.assertEqual(True, all_true)

    def test_scale_coords_folder(self):

        input_path = os.path.join(os.path.dirname(__file__),
                            "../resources/coords_data/")
        output_path = os.path.join(os.path.dirname(__file__),
                                  "../resources/out/")
        scale.scale(input_path=input_path,
                    output_path=output_path,
                    scale_factor=0.5)


        written_file_path = os.path.join(os.path.dirname(__file__),
                                   "../resources/out/example.coords")
        input_coords = np.atleast_2d(np.genfromtxt(os.path.join(input_path,"example.coords")))
        written_coords = np.atleast_2d(np.genfromtxt(written_file_path))

        scal_fact_mat = 2*written_coords-input_coords

        scal_fact_mat = np.abs(scal_fact_mat)<10**-4
        all_true = np.all(scal_fact_mat)
        os.remove(written_file_path)
        self.assertEqual(True, all_true)

    def test_rotate_in_volume(self):

        coords = np.array([
            [0,0,0],
            [10,0,0],
            [0,10,0],
            [0,0,10],
            [1,1,1]
        ])
        exp_result = np.array([
            [0,0,0],
            [1,1,1]
        ])
        shape = np.array([5,5,5])
        res = rotatecboxcoords.RotateCBOXCoords.remove_out_of_volume_coords(coords,shape)
        np.testing.assert_array_equal(exp_result, res)

    def test_file_type_coords(self):
        input_path = os.path.join(os.path.dirname(__file__),
                                  "../resources/coords_data/example.coords")
        filetype = boxmanager_tools.get_file_type(input_path)
        self.assertEqual(filetype,boxmanager_tools.TYPE_COORDS)

    def test_file_type_helicon(self):
        input_path = os.path.join(os.path.dirname(__file__),
                                  "../resources/segmented_fila_data/segmented_filament.box")
        filetype = boxmanager_tools.get_file_type(input_path)
        self.assertEqual(filetype,boxmanager_tools.TYPE_EMAN_HELICON)

    def test_file_type_cbox(self):
        input_path = os.path.join(os.path.dirname(__file__),
                                  "../resources/cbox_data/example3D_filament.cbox")
        filetype = boxmanager_tools.get_file_type(input_path)
        self.assertEqual(filetype,boxmanager_tools.TYPE_CBOX)


if __name__ == '__main__':
    unittest.main()





from cryoloBM.bmtool import BMTool
from argparse import ArgumentParser
import argparse
from typing import List
import numpy as np
from cryolo.utils import Filament, BoundBox
from cryolo.CoordsIO import read_cbox_boxfile
from glob import glob
import os

class CBOX2Coords(BMTool):

    def get_command_name(self) -> str:
        '''
        :return: Command name
        '''
        return "cbox2coords"

    @staticmethod
    def generate_coords_array_filaments(data: List[Filament]):
        all_fil_coords = []
        for i, fil in enumerate(data):
            coords = CBOX2Coords.generate_coords_array(fil.boxes)
            id = np.array([i]*len(fil.boxes))
            id = id.reshape((len(id),1))
            coords_fid = np.append(coords, id, axis=1)
            all_fil_coords.append(coords_fid)
        return np.concatenate(all_fil_coords)

    @staticmethod
    def generate_coords_array(data: List[BoundBox]):
        coords_dat = np.empty(shape=(len(data), 3))
        for i, box in enumerate(data):
            coords_dat[i, 0] = box.x + box.w/2
            coords_dat[i, 1] = box.y + box.h/2
            coords_dat[i, 2] = box.z
        return coords_dat

    def create_parser(self, parentparser : ArgumentParser) -> ArgumentParser:
        '''
        :param parentparser: ArgumentPaser where the subparser for this tool needs to be added.
        :return: Argument parser that was added to the parentparser
        '''
        parser = parentparser.add_parser(
            self.get_command_name(),
            help="Converts CBOX files to .coords files.",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )

        req = parser.add_argument_group(
            "Arguments",
            "Converts CBOX files to .coords files",
        )

        req.add_argument(
            "-i",
            "--input",
            required=True,
            help="Path to folder with CBOX files ",
        )

        req.add_argument(
            "-o",
            "--out",
            required=True,
            help="Output path",
        )

        return parser


    def run(self, args):
        input_path = args.input
        output_path = args.out
        cbox_files = glob(os.path.join(input_path,"*.cbox"))


        for cbox_file in cbox_files:
            cbox_dat = read_cbox_boxfile(cbox_file)
            is_filament = isinstance(cbox_dat[0], Filament)
            filename = os.path.splitext(os.path.basename(cbox_file))[0]

            if is_filament:
                coords_dat = CBOX2Coords.generate_coords_array_filaments(cbox_dat)
            else:
                coords_dat = CBOX2Coords.generate_coords_array(cbox_dat)

            out_coords = os.path.join(output_path, "COORDS/")
            os.makedirs(out_coords, exist_ok=True)
            out_file_pth = os.path.join(out_coords, filename + ".coords")
            np.savetxt(out_file_pth, coords_dat[:, :3], fmt='%f')

            if is_filament:
                out_coords = os.path.join(output_path, "COORDS_FID/")
                os.makedirs(out_coords, exist_ok=True)
                out_file_pth = os.path.join(out_coords, filename + ".coords")
                np.savetxt(out_file_pth, coords_dat, fmt='%f')


"""
In this tool we use the Model View Controller (MVC) design pattern.
The MCV can be considered an approach to distinguish between the data model, processing control and the user interface.
It neatly separates the graphical interface displayed to the user from the code that manages the user actions.

In this file we define the 'Boxmanager_controller' which has the logic that binds model (boxmanager_model.py) and view
(boxmanager_view.py). Accepts input and converts it to commands for the model or view

Except for the 'run' function all the other function are just calling the function defined in the helper files
"""

from sys import exit as sys_exit
from os import path
from cryoloBM import logic_dropdown_menu, helper_image, import_data, helper_tab_thresholding, helper_tab_tracing,writer,constants,helper_GUI,helper
from cryoloBM import boxmanager_model, boxmanager_view   # for type hinting purpose

class Boxmanager_controller:
    def __init__(self, model:boxmanager_model, view:boxmanager_view, app:helper_GUI.QtG)->None:
        """
        :param model: Boxmanager_model obj
        :param view: Boxmanager_view obj
        :param app: QTWidgets object
        """
        self.model = model
        self.view = view
        self.app = app


    def get_current_filename(self, with_extension:bool = False)->None:
        """
        Return only the filename of the current_image_path. With or without extension.
        Typically with extension for setting the window's title
        """
        if not self.model.current_image_path:
            return None
        f = path.splitext(path.basename(self.model.current_image_path))
        return f[0]+f[1]  if with_extension else f[0]


    def run(self)->None:
        """
        This function run the boxmanager
        """
        self.view.setup(self)

        # manage the run via command line
        if self.model.image_folder_path is not None:
            self.view.image_folder_path = self.model.image_folder_path
            if self.model.is_tomo:
                if path.isdir(self.model.image_folder_path):
                    self.open_tomo_folder(via_cli= True)
                else:
                    helper.set_current_image_path(controller=self,current_image_path= self.model.image_folder_path)
                    self.open_tomo_image(via_cli=True)
            else:
                if path.isdir(self.model.image_folder_path):
                    self.open_SPA_folder(via_cli= True)
                else:
                    helper.set_current_image_path(controller=self, current_image_path=self.model.image_folder_path)
                    self.open_SPA_image(via_cli=True)
            if self.model.box_dir:
                if self.model.type_case == constants.Type_case.SPA.value:
                    import_data.import_SPA(controller=self, box_dir=self.model.box_dir, keep=False)
                elif self.model.type_case == constants.Type_case.TOMO_FOLDER.value:
                    import_data.import_tomo_folder(controller=self, box_dir=self.model.box_dir, keep=False)
                else:
                    import_data.import_single_tomo(controller=self, path_3d_file=path.splitext(self.model.box_dir)[0],index_file=None,keep=False)



                # update counters
                self.view.update_tree_boxsizes(update_current=False, all_tomos=True)
                self.view.update_all_global_counters()

                #set cbox default option
                self.view.is_updating_params = True
                kind_of_cbox = self.model.is_cbox or self.model.is_cbox_untraced
                if kind_of_cbox:
                    self.view.set_cbox_visualization(has_filament=self.model.has_filament)
                self.view.thresholding_tab_blurring(is_enable=kind_of_cbox,
                                                     kind_of_cbox=kind_of_cbox, has_filament=self.model.has_filament,
                                                     type_case=self.model.type_case)
                self.view.tracing_tab_blurring(is_enable=self.model.is_cbox_untraced,
                                                     has_filament=self.model.has_filament)
                self.view.blur_picking_combobox(has_filament=self.model.has_filament, is_importing=True)


                if self.model.box_dictionary.has_filament:
                    self.model.box_dictionary.fil_start_end_vis = False
                    self.view.set_cbox_visualization(has_filament=self.model.has_filament)

                # in case of cbox filament some options are not available
                if self.model.has_filament:
                    self.view.blur_estimated_size()
                    self.model.box_distance_filament = self.model.box_dictionary.set_box_distance_dict_after_import()
                    self.view.line_setText(line=self.view.box_distance_filament_line, value=str(self.model.box_distance_filament))

            self.view.is_updating_params = False

        sys_exit(self.app.exec_())

    def close_boxmanager(self)->None:
        """
        It is called via File->close
        Close the boxmanager script.
        """
        logic_dropdown_menu.close_boxmanager(self)

    def reset_config(self)->None:
        """
        It is called via File->Reset
        Reset all the variables of the model and restore the GUI starting layout
        """
        logic_dropdown_menu.reset_config(self)

    def import_box_files(self)->None:
        """
        Function to import box files from file
        """
        import_data.import_data(self)

    def save_on_files(self)->None:
        """
        Function to to save the results on file. It write all the possible formats.
        It was write_all_type in the old version
        """
        writer.save_on_files(self)

    def open_SPA_folder(self, via_cli:bool = False)->None:
        """
        Function to read micrograph's images from a folder
        :param via_cli: True if we run from command line
        """
        logic_dropdown_menu.open_SPA_folder(self, via_cli = via_cli)

    def open_SPA_image(self, via_cli:bool = False)->None:
        """
        Function to read a single micrograph's image
        :param via_cli: True if we run from command line
        """
        logic_dropdown_menu.open_SPA_image(self, via_cli = via_cli)

    def open_tomo_folder(self, via_cli:bool = False)->None:
        """
        Function to read tomo's images from a folder
        :param via_cli: True if we run from command line
        """
        logic_dropdown_menu.open_tomo_folder(self, via_cli = via_cli)
        self.model.is_cbox_untraced = True

    def open_tomo_image(self, via_cli:bool = False)->None:
        """
        Function to read a single tomo's image
        :param via_cli: True if we run from command line
        """
        logic_dropdown_menu.open_tomo_image(self, via_cli = via_cli)
        self.model.is_cbox_untraced = True

    def show_size_distribution(self)->None:
        """
        Function to show the size distribution
        """
        logic_dropdown_menu.show_size_distribution(self)

    def show_confidence_histogram(self)->None:
        """
        Function to to show the confidence histogram
        """
        logic_dropdown_menu.show_confidence_histogram(self)

    def event_image_changed(self, root_tree_item:helper_GUI.QtG)->None:
        """
        Manage the 'image changes' event
        :param root_tree_item: It is filled in automatic via PyQT
        """
        helper_image.event_image_changed(self, root_tree_item=root_tree_item)

    def onclick(self, event:helper_GUI.plt_qtbackend)->None:
        """
        It is called when we click on the image
        :param event: the click event
        """
        helper_image.onclick(self,event=event)

    def onmove(self, event:helper_GUI.plt_qtbackend)->None:
        """
        It is called when we click on the image
        :param event: the click event
        """
        helper_image.onmove(self, event=event)

    def onresize(self, event:helper_GUI.plt_qtbackend)->None:
        """
        It is called when we resize the main window
        :param event: the click event
        """
        helper_image.onresize(self,event=event)

    def ondraw(self, event:helper_GUI.plt_qtbackend)->None:
        """
        It is called when we use the toolbar
        :param event: the click event
        """
        helper_image.ondraw(self,event=event)

    def onrelease(self, event:helper_GUI.plt_qtbackend)->None:
        """
        It is called when we release the mouse button
        :param event: the click event
        """
        helper_image.onrelease(self, event=event)

    def myKeyPressEvent(self, event:helper_GUI.plt_qtbackend)->None:
        """
        It is called when we press the keyboard
        :param event: the press keyboard event
        """
        helper_image.myKeyPressEvent(self, event=event)

    def box_size_changed(self)->None:
        """
        It changes the value of the "box size" filling the 'box size' line in the visualization tab.
        It is triggered pressing enter (when the cursor is into the line) or button 'set'
        """
        helper_image.box_size_changed(self)

    def box_distance_filament_changed(self):
        """
        It changes the value of the "Box distance" filling the 'Box distance' line in the visualization tab.
        It is triggered pressing enter (when the cursor is into the line) or button 'set'
        """
        helper_image.box_distance_filament_changed(self)

    def use_estimated_size_changed(self)->None:
        """
        It changes the value of the "box size" filling the 'box size' line in the visualization tab.
        It is usable only after loading data from cbox file. It gets the new "box size" from these files
        Pay attention: each particle will have its own estimated size
        It is triggered pressing enter (when the cursor is into the line) or button 'set'
        """
        # otherwise it is triggered when resetting
        if not self.view.is_updating_params:
            helper_image.use_estimated_size_changed(self)


    def visualization_changed(self)->None:
        """
        It changes the value of the "visualization_combobox" changing the 'visualization' combobox in the visualization tab.
        It is triggered when the combobox is changed
        """
        if not self.view.is_updating_cb:
            helper_image.visualization_changed(self)

    # FILTERING TAB OPTION
    def apply_filter(self)->None:
        """
        It applies the lowpass filter present in filtering tab.
        """
        helper_image.apply_filter(self)

    # THRESHOLDING TAB OPTION
    def conf_thresh_changed(self)->None:
        """
        It is called when the 'confidence threshold', present in thresholding tab, is changed
        """
        helper_tab_thresholding.conf_thresh_changed(self)

    def conf_thresh_label_changed(self)->None:
        """
        It is called when the label of the 'confidence threshold', present in thresholding tab, is changed
        """
        helper_tab_thresholding.conf_thresh_label_changed(self)

    def lower_size_thresh_changed(self)->None:
        """
        It is called when the 'lower size of the confidence threshold', present in thresholding tab, is changed
        """
        helper_tab_thresholding.lower_size_thresh_changed(self)

    def lower_size_label_changed(self)->None:
        """
        It is called when the label of 'lower size of the confidence threshold', present in thresholding tab, is changed
        """
        helper_tab_thresholding.lower_size_label_changed(self)

    def upper_size_thresh_changed(self)->None:
        """
        It is called when the 'upper size of the confidence threshold', present in thresholding tab, is changed
        """
        helper_tab_thresholding.upper_size_thresh_changed(self)

    def upper_size_label_changed(self)->None:
        """
        It is called when the label of 'upper size of the confidence threshold', present in thresholding tab, is changed
        """
        helper_tab_thresholding.upper_size_label_changed(self)

    # TRACING OPTION
    def searchRange_changed(self)->None:
        """
        It is called when the 'search range', present in tracing tab, is changed
        """
        helper_tab_tracing.searchRange_changed(self)

    def searchRange_label_changed(self)->None:
        """
        It is called when the label of the 'search range', present in tracing tab, is changed
        """
        helper_tab_tracing.searchRange_label_changed(self)

    def memory_changed(self)->None:
        """
        It is called when the 'memory', present in tracing tab, is changed
        """
        helper_tab_tracing.memory_changed(self)

    def memory_label_changed(self)->None:
        """
        It is called when the label of the 'memory', present in tracing tab, is changed
        """
        helper_tab_tracing.memory_label_changed(self)

    def min_length_changed(self)->None:
        """
        It is called when the 'min lenght', present in tracing tab, is changed
        """
        helper_tab_tracing.min_length_changed(self)

    def min_length_label_changed(self)->None:
        """
        It is called when the label of the 'min lenght', present in tracing tab, is changed
        """
        helper_tab_tracing.min_length_label_changed(self)

    def min_edge_weight_changed(self)->None:
        """
        It is called when the 'min edge weight', present in tracing tab, is changed
        """
        helper_tab_tracing.min_edge_weight_changed(self)

    def min_edge_weight_label_changed(self)->None:
        """
        It is called when the label of the 'min edge weight', present in tracing tab, is changed
        """
        helper_tab_tracing.min_edge_weight_label_changed(self)

    def win_size_changed(self)->None:
        """
        It is called when the 'win size', present in tracing tab, is changed
        """
        helper_tab_tracing.win_size_changed(self)

    def win_size_label_changed(self)->None:
        """
        It is called when the label of the 'win size', present in tracing tab, is changed
        """
        helper_tab_tracing.win_size_label_changed(self)

    def preview(self)->None:
        """
        it is called when the tracing is done and show the image with the tracing results
        """
        helper_tab_tracing.preview(self)

    def trace(self)->None:
        """
        It starts the tracing
        """
        helper_tab_tracing.trace(self)

    def set_has_filament(self, has_filament:bool)->None:
        """
        Set the 'has_filament' internal vars
        :param has_filament: True if we use filaments. Otherwise False
        """
        self.model.has_filament = has_filament
        if self.model.box_dictionary:
            self.model.box_dictionary.has_filament = has_filament


"""
It this file are defined all the constant values used in the project
"""

from enum import Enum

# unique id used to identify the center of a filamente during the picking phase
CENTER_LINE_ID = "invalid"

class Writer_filament_suffix(Enum):
    """
    List of suffix of output folder in the filament cases
    """
    HELICON_SUFFIX = "_HELICON"
    CBOX_FILAMENT_SUFFIX = "_FILAMENT"
    EMAN_STAR_FILAMENT_SUFFIX = "_START_END"


class Slider_release_fcall(Enum):
    """
    List of function call to map in changed_slider_release to run the related controller's function
    the boxmanager main win
    """
    CONF = 0
    LOWER_SIZE = 1
    UPPER_SIZE = 2
    # num_boxes , last thresholding option
    SEARCH_RANGE = 4
    MEMORY = 5
    MIN_LENGTH = 6
    MIN_EDGE_W = 7
    WIN_SIZE = 8


class Picking_cb(Enum):
    """
    List of possible picking choices present in the 'picking' combobox in the boxmanager main win
    """
    PARTICLE = 'Particle'
    FILAMENT = 'Filament'

class Display_visualization_cb(Enum):
    """
    For setting the correct option in the visualization combobox (see class Visualization_cb)
    """
    PARTICLE = 0        # enable (CIRCLE,RECT)
    FILAMENT_2D = 1     # enable (CIRCLE_SEGMENTED, RECT_FILAMENT_SEGMENTED, RECT_FILAMENT_START_END)
    FILAMENT_3D = 2     # enable (CIRCLE_SEGMENTED, RECT_FILAMENT_SEGMENTED)

class Visualization_cb(Enum):
    """
    List of possible visualization choices present in the 'visualization'' combobox in the 'visualization' tab in
    the boxmanager main win
    """
    CIRCLE = 'Circle'
    RECT = 'Rectangle'

    # value for the filament visualization
    CIRCLE_SEGMENTED = 'Circle (Filament Segmented)'  # list of cirlces
    RECT_FILAMENT_SEGMENTED = 'Rectangle (Filament Segmented)'  # list of rectangles
    RECT_FILAMENT_START_END = 'Rectangle (Filament Start-end)'  # single rectangle not enable in 3D

class Type_case(Enum):
    """
    Identify if the loaded case is SPA, single tomo or folder of tomo
    """
    NOT_LOADED = None
    SPA = 0
    TOMO = 1
    TOMO_FOLDER = 2

class Filetype_name(Enum):
    """
    Map the filetype with values used to read the image via 'helper.read_image'
    """
    JPG = 0
    JPEG = 0
    TIF = 1
    TIFF = 1
    MRC = 2
    MRCS = 2
    REC =2


class Default_settings_value(Enum):
    """
    List of the default setting values
    """
    # default value for the picking combobox not present in any tabs
    PICKING_CB = Picking_cb.PARTICLE.value

    # default value for the 'Visualization' tab
    DEFAULT_BOX_SIZE = 200
    DEFAULT_BOX_DISTANCE_FILAMENT_SIZE = int(DEFAULT_BOX_SIZE / 10)
    VISUALIZATION_CB = Visualization_cb.CIRCLE.value

    # default value for the 'Thresholding' tab
    DEFAULT_UPPER_SIZE_THRESH = 99999
    DEFAULT_LOWER_SIZE_THRESH = 0
    DEFAULT_CURRENT_CONF_THRESH = 0.3
    DEFAULT_CURRENT_NUM_BOXES_THRESH = 0
    DEFAULT_MIN_NUM_BOXES_THRESH = 0
    DEFAULT_MAX_NUM_BOXES_THRESH = 100
    DEFAULT_MIN_EDGE_WEIGHT = 0.4

    # default value for the 'Filtering' tab
    DEFAULT_FILTER_FREQ = 0.1

    # default value for the 'Tracing' tab
    DEFAULT_SEARCH_RANGE = 5
    DEFAULT_MEMORY = 0
    DEFAULT_MIN_LENGTH = 5



"""
In this file the 'MySketch' class is defined.
This class is used for saving all the data of a sketch
"""

from matplotlib.patches import  Rectangle,Circle
from matplotlib import colors
from cryolo import utils
import numpy as np
from typing import Union,Tuple,Dict

class MySketch:
    def __init__(self, xy: Union[Tuple[float, float], None] = None, width: Union[int, None] = None,
                 height: Union[int, None] = None, is_tomo: bool = False, angle: float = 0.0,
                 est_size: Union[int, None] = None, confidence: Union[float, None] = None,
                 only_3D_visualization: bool = False, num_boxes: int = 1,
                 meta: Union[Dict[str, Tuple[float, float]], None] = None, z: Union[int, None] = None,
                 id: Union[int, None] = None, **kwargs) -> None:
        """
        :param xy: tuple representing the bottom and left rectangle coordinates. In according with matplotlib.patches.py class Rectangle
        :param width: of the rect
        :param height: of the rect
        :param is_tomo: True if the sketch will belong to a tomography
        :param angle: rotation in degrees anti-clockwise about *xy* (default is 0.0). In according with matplotlib.patches.py class Rectangle
        :param est_size: size estimated by crYOLO ( .cbox cases)
        :param confidence: confidence value estimated by crYOLO
        :param only_3D_visualization: If true this box is not a real box picked or detected by crYOLO but it is created in the 3D visualization feature
        :param num_boxes: number of boxes of the 3D particle (CBOX_3D file). in case of 2D it has value 1
        :param meta: meta data values of a Bbox instance
        :param z: slice in tomography
        :param id: particle id. I set it ONLY when i create the 3D visualization for detecting all the 2D part of the 'id 3D particle

        NB:
        in case of particle a rectangle will be a square --> box_size = height = width (hence radius = width/2)
        In case of filament it will be NOT A SQUARE, we have to decide how manage some stuff BUT the radius = It would be size perpendicular to the filament
        """

        # sketch id. The idea is to use the value for managing the pick and delete options in the 3D visualization
        self.id=id
        self.only_3D_visualization=only_3D_visualization
        self.meta = meta

        #it is the number of boxes of the 3D particle (CBOX_3D file). in case of 2D it has value 1
        # NB:
        #   the 'current_num_boxes_thresh' is set 0 for default, that means that in helper.check_if_should_be_visible
        #   2D particle (hence whose in the untraced_cbox files too) will be always visible when check 'box.num_boxes > num_boxes_thresh'
        self.num_boxes = num_boxes

        """
        Value used , together with meta["_filamentid"]for the filament tracking
        z = index tomo
        """
        self.z = z

        self.sketch_circle = MyCircle(xy=(xy[0]+width/2,xy[1]+width/2), radius=int(width / 2), is_tomo= is_tomo, est_size=est_size, confidence = confidence, **kwargs)
        self.sketch_rect = MyRectangle(xy=xy, width=width, height=height, is_tomo= is_tomo, angle=angle, est_size=est_size, confidence = confidence, **kwargs)

    def get_as_BBox(self, confidence:float = None, z:int = None)->utils.BoundBox:
        """
        convert the sketch in a cryolo Bounding box and return it
        :param confidence: confidence value. 1 if the bbox will pass to cryolo.grouping3d.do_tracing
        :param z: slice of the 3D tomo, None otherwise
        return: a BBox
        """

        """
            NB:
                The x,y of a BBox are the center of the box
                The x,y of a circle_sketch are the center of the circle
                The x,y of a rect_sketch are the bottom and left rectangle coordinates
        """
        w = self.get_width()

        #The bbox are rectangles hence i need the x,y values of the rectangles sketch.
        xy = self.get_xy(circle=False)

        #depth is always 1 even in the tomo because the box are on a slice, hence in 2D hence depth=1
        b= utils.BoundBox(x=xy[0], y=xy[1], w=w, h=w, c=confidence, classes=["particle"], z=z, depth=1)
        meta = (w,w)
        if self.meta is not None and "est_box_size" in self.meta:
            meta = self.meta["est_box_size"]
        b.meta = {"boxsize_estimated": meta}       # i need it in the 'cryolo.grouping3d.convert_traces_to_bounding_boxes'
        return b

    def resize(self, new_size:int)->None:
        confidence = self.get_confidence()
        is_tomo =self.get_is_tomo()
        color = self.getSketch().get_edgecolor()

        """est_size contained the estimated size of cryolo, when loaded from '.cbox' and should be never changed"""
        est_size = self.get_est_size()

        self.remove_instances()

        self.sketch_circle = MyCircle(xy=self.get_xy(circle=True), radius=new_size/2, is_tomo=is_tomo, est_size=est_size,
                                      confidence=confidence, linewidth=1, edgecolor=color, facecolor="none")

        xy = self.get_xy(circle=False)
        x = xy[0] + (self.get_width()-new_size)/2
        y = xy[1] + (self.get_width()-new_size)/2
        self.sketch_rect = MyRectangle(xy=(x,y), width=new_size, height=new_size, is_tomo=is_tomo,
                                       angle=self.get_angle(), est_size=est_size, confidence=confidence,
                                       linewidth=1, edgecolor=color, facecolor="none")
        self.createSketches()  # create the new instances

    def remove_instances(self)->None:
        """
        Remove the instances of the sketches
        """
        self.sketch_circle.remove_istance()
        self.sketch_rect.remove_istance()

    def createSketches(self)->None:
        """
        Create the instances of the sketches
        """
        self.getSketch(circle =True)
        self.getSketch(circle =False)

    def getSketch(self, circle:bool = False)->Union[Rectangle,Circle]:
        """
        Return an matplotlib instance of my sketch
        """
        if circle:
            return self.sketch_circle.getSketch()
        return self.sketch_rect.getSketch()

    def set_Sketches_visible(self, visible:bool =True)->None:
        self.getSketch(circle=True).set_visible(visible)
        self.getSketch(circle=False).set_visible(visible)

    def set_xy(self, xy:Tuple[float,float], circle:bool = False)->None:
        if circle:
            self.sketch_circle.xy = xy
        else:
            self.sketch_rect.xy = xy

    def set_radius(self, radius:int)->None:
        self.sketch_circle.radius = radius

    def set_width(self, width:int)->None:
        self.sketch_rect.width = width

    def set_height(self, height:int)->None:
        self.sketch_rect.set_height(new_height= height)

    def set_is_tomo(self, is_tomo:bool)->None:
        self.sketch_rect.is_tomo = is_tomo
        self.sketch_circle.is_tomo = is_tomo

    def set_angle(self, angle:float)->None:
        self.sketch_rect.angle = angle

    def set_est_size(self, est_size:int)->None:
        self.sketch_rect.est_size = est_size
        self.sketch_circle.est_size = est_size

    def set_confidence(self, confidence:float)->None:
        self.sketch_rect.confidence = confidence
        self.sketch_circle.confidence = confidence

    def set_colour(self, colour:str)->None:
        self.sketch_rect.kwargs["edgecolor"]=colour
        self.sketch_circle.kwargs["edgecolor"] = colour

    def get_xy(self, circle:bool = False)->Tuple[float,float]:
        if circle:
            return self.sketch_circle.xy
        return self.sketch_rect.xy

    def get_radius(self)->float:
        return self.sketch_circle.radius

    def get_width(self)->int:
        return self.sketch_rect.width

    def get_height(self)->int:
        return self.sketch_rect.height

    def get_is_tomo(self)->bool:
        return self.sketch_rect.is_tomo

    def get_angle(self)->float:
        return self.sketch_rect.angle

    def get_est_size(self, circle:bool = False)->int:
        if circle:
            return self.sketch_circle.est_size
        return self.sketch_rect.est_size

    def get_confidence(self)->float:
        return self.sketch_rect.confidence

    def get_color(self)->Tuple[float,float,float,float]:
        return self.getSketch().get_edgecolor()


class MyCircle:
    def __init__(self, xy: Union[Tuple[float, float], None], radius:float, is_tomo:bool = False, est_size:Union[int, None]=None, confidence:Union[float, None] = None, **kwargs)->None:
        self.confidence = confidence
        self.est_size = est_size
        self.is_tomo = is_tomo
        self.xy = xy
        self.radius = radius
        self.circleInstance = None
        self.kwargs = kwargs

    def getSketch(self)->Circle:
        if self.circleInstance is None:
            self.circleInstance = Circle(xy=self.xy, radius=self.radius, **self.kwargs)
        return self.circleInstance

    def remove_istance(self)->None:
        self.circleInstance = None


class MyRectangle:
    def __init__(self, xy: Union[Tuple[float, float], None], width:int, height:int, is_tomo:bool = False, angle:float=0.0, est_size:Union[int, None]=None, confidence:Union[float, None] = None, **kwargs)->None:
        self.confidence = confidence
        self.est_size = est_size
        self.is_tomo = is_tomo
        self.xy = xy
        self.width = width
        self.height = height
        self.angle = angle # degree
        self.rectInstance = None
        self.kwargs = kwargs

    def getSketch(self)->Rectangle:
        if self.rectInstance is None:
            self.rectInstance = Rectangle(self.xy, self.width, self.height, self.angle, **self.kwargs)
        return self.rectInstance

    def set_height(self, new_height):
        """Set the height. This method should update also xy when the height is changed"""
        angle = self.angle%180
        sign = 1
        if angle >= 90:
            sign = -1
        rad_angle = np.deg2rad(angle)

        x_fin = np.cos(rad_angle)
        y_fin = np.sin(rad_angle)
        help = x_fin
        x_fin = y_fin
        y_fin = -1*help
        orth_vect = np.array([x_fin,y_fin])

        old_xy = np.array(self.xy)

        new_xy = old_xy + sign*orth_vect*(new_height-self.height)/2

        self.xy = (new_xy[0],new_xy[1])
        self.height = new_height


    def remove_istance(self)->None:
        self.rectInstance = None


import os
import glob
import numpy as np
import pandas as pd
from pyStarDB import sp_pystardb as star
from cryoloBM.bmtool import BMTool
import argparse
import tqdm

from argparse import ArgumentParser

class Coords2WarpTool(BMTool):

    def get_command_name(self) -> str:
        return "coords2star"

    def create_parser(self, parser) -> ArgumentParser:

        parser_coords2warp = parser.add_parser(
            self.get_command_name(),
            help="Converts tomo coords file to Warp/Relion4 compatible star file.",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )

        c2w_required_group = parser_coords2warp.add_argument_group(
            "Required arguments",
            "Converts .coords to a WARP compatible .star file.",
        )

        c2w_required_group.add_argument(
            "-i",
            "--input",
            required=True,
            help="Input folder or file. Files should be coords format. If the .coords provides a filament id, prior information are automatically added.",
        )

        c2w_required_group.add_argument(
            "-o",
            "--output",
            required=True,
            help="Output folder where to write the WARP compatible star file.",
        )

        c2w_required_group.add_argument(
            "--scale",
            type=float,
            default=1.0,
            help="Coordinates get scaled by this factor. This is useful incase you have .coords files from binned reconstructions.",
        )

        c2w_required_group.add_argument(
            "--apix",
            required=False,
            default=None,
            type=float,
            help="Pixel size in Angstrom after scaling. This value filled into the _rlnDetectorPixelSize column within the star file.",
        )

        c2w_required_group.add_argument(
            "--mag",
            type=float,
            default=None,
            help="Magnification",
        )

        c2w_required_group.add_argument(
            "--flipratio",
            type=float,
            default=0,
            help="Only relevant for filaments. This value is filled into the _rlnAnglePsiFlipRatio column within the star file.",
        )

        c2w_required_group.add_argument(
            "--relion4",
            action='store_true',
            default=False,
            help="Flat if relion4 style should be used",
        )

        return parser_coords2warp

    def run(self, args):

        all_relion_data = convert(
            input_path=args.input,
            scale=args.scale,
            pixelsize=args.apix,
            magnification=args.mag,
            flip_ratio=args.flipratio,
            warp_style=not args.relion4
        )

        if len(all_relion_data) > 0:
            postfix = "warp"
            if args.relion4:
                postfix="relion4"
            os.makedirs(args.output, exist_ok=True)
            df = pd.concat(all_relion_data)
            output_file_path = os.path.join(args.output, f"particles_{postfix}.star")
            sfile = star.StarFile(output_file_path)
            sfile.update('', df, True)
            sfile.write_star_file(overwrite=True)
            return df
        return None

def calang(current_point,previous_point, xyz_indices = [0,1,2]):
    '''
    Calculate the the tilt and psi angle of a certain point based on its coords and the point before it.
    :param current_point:
    :param previous_point:
    :return: tilt and psi angle
    '''

    p_x = float(current_point[xyz_indices[0]])
    p_y = float(current_point[xyz_indices[1]])
    p_z = float(current_point[xyz_indices[2]])
    pminus_x = float(previous_point[xyz_indices[0]])
    pminus_y = float(previous_point[xyz_indices[1]])
    pminus_z = float(previous_point[xyz_indices[2]])

    vector = [p_x-pminus_x,p_y-pminus_y,p_z-pminus_z]
    psi = -np.arctan2(vector[1],vector[0])*180/np.pi
    xylength = np.sqrt(vector[0]**2+vector[1]**2)
    tilt = -np.arctan2(xylength,vector[2])*180/np.pi
    tilt = tilt+180

    return tilt, psi

def add_prior_information(rlndata):
    xindex, yindex, zindex, tilt_index, angle_index, id_index = [rlndata.columns.get_loc(c) for c in ['_rlnCoordinateX', '_rlnCoordinateY', '_rlnCoordinateZ', '_rlnAngleTiltPrior', '_rlnAnglePsiPrior', '_rlnHelicalTubeID']]
    npdata = rlndata.to_numpy()
    filids = np.unique(npdata[:,id_index])

    for filid in filids:
        filmask = npdata[:, id_index] == filid
        rlndata_fil = npdata[filmask]
        if len(rlndata_fil) < 2:
            print("Filament with ID", filid, "has length", len(rlndata_fil), "and is therefore ignored" )
            continue
        for row in range(rlndata_fil.shape[0]):
            if row == 0:
                tilt, psi = calang(rlndata_fil[1,:],rlndata_fil[row,:], xyz_indices=[xindex,yindex,zindex])
            elif row == (rlndata_fil.shape[0]-1):
                tilt, psi = calang(rlndata_fil[row,:],rlndata_fil[-2,:], xyz_indices=[xindex,yindex,zindex])
            else:
                tilt_1, psi_1 = calang(rlndata_fil[row,:],rlndata_fil[row-1,:], xyz_indices=[xindex,yindex,zindex])
                tilt_2, psi_2 = calang(rlndata_fil[row+1, :], rlndata_fil[row, :], xyz_indices=[xindex,yindex,zindex])
                tilt = np.mean([tilt_1,tilt_2])
                psi = np.mean([psi_1,psi_2])

            rlndata_fil[row, tilt_index] = tilt
            rlndata_fil[row, angle_index] = psi
        npdata[filmask] = rlndata_fil
    return npdata

def convert(
        input_path: str,
        pixelsize: float,
        magnification: int,
        scale: float,
        flip_ratio: float =0,
        warp_style = True) -> pd.DataFrame:
    '''
    :param input_path: Input path with .coords files
    :param output_path: Path to folder where results should be written
    :param pixelsize: Pixel size in angstrom
    :param magnification: Magnification value
    :param scale: All coordaintes get scaled by this factor
    :param flip_ratio: value that is used for flip ratio
    :return: None
    '''
    if os.path.isfile(input_path):
        files = [input_path]
    else:
        path = os.path.join(os.path.abspath(input_path), "*.coords")
        files = glob.glob(path)
    all_relion_data = []
    for pth in tqdm.tqdm(files, desc="Converting"):
        coords = np.atleast_2d(np.genfromtxt(pth))
        if coords.size == 0:
            print(pth,"is empty. Skip.")
            continue
        has_fid = coords.shape[1]==4
        micrographname = os.path.splitext(os.path.basename(pth))[0]
        if has_fid:
            micrographname = micrographname[:-4] # remove _fid

        if warp_style:
            micrographname = micrographname + ".tomostar"
            first_column = '_rlnMicrographName'
        else:
            micrographname = micrographname + ".star"
            first_column = '_rlnTomoName'
        columns = [first_column, '_rlnCoordinateX', '_rlnCoordinateY', '_rlnCoordinateZ']
        if magnification:
            columns.append('_rlnMagnification')
        if pixelsize:
            columns.append('_rlnDetectorPixelSize')

        if has_fid:
            columns.append('_rlnHelicalTubeID')
            columns.append('_rlnAngleTiltPrior')
            columns.append('_rlnAnglePsiPrior')
            columns.append('_rlnAnglePsiFlipRatio')

        rlndata = np.zeros(shape=(coords.shape[0], len(columns)))

        for i, row in enumerate(coords):
            rlndata[i, columns.index("_rlnCoordinateX")] = float(row[0])*scale
            rlndata[i, columns.index("_rlnCoordinateY")] = float(row[1])*scale
            rlndata[i, columns.index("_rlnCoordinateZ")] = float(row[2])*scale
            if magnification:
                rlndata[i, columns.index("_rlnMagnification")] = magnification
            if pixelsize:
                rlndata[i, columns.index("_rlnDetectorPixelSize")] = pixelsize
            if has_fid:
                rlndata[i, columns.index("_rlnHelicalTubeID")] = row[3]



        df = pd.DataFrame(rlndata, columns=columns)
        if has_fid:
            # Calculate and add prio
            add_prior_information(df)

            # Set _rlnAnglePsiFlipRatio #

            df['_rlnAnglePsiFlipRatio'] = flip_ratio

        df.iloc[:,0] = micrographname
        if not warp_style:
            df = df.drop('_rlnHelicalTubeID', axis=1)
        all_relion_data.append(df)


    return all_relion_data



from cryoloBM.bmtool import BMTool
from argparse import ArgumentParser
import argparse
from typing import List
import numpy as np
from cryolo.utils import Filament, BoundBox
from cryolo.CoordsIO import read_cbox_boxfile
from glob import glob
from cryolo.CoordsIO import write_cbox_file

import os

class Coords2CBOX(BMTool):

    def get_command_name(self) -> str:
        '''
        :return: Command name
        '''
        return "coords2cbox"

    def create_parser(self, parentparser: ArgumentParser) -> ArgumentParser:
        helptxt = "Converts .coords files to .cbox files."
        parser = parentparser.add_parser(
            self.get_command_name(),
            help=helptxt,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )

        req = parser.add_argument_group(
            "Arguments",
            helptxt,
        )

        req.add_argument(
            "-i",
            "--input",
            required=True,
            help="Path to folder with .coords files / path to single .coords file ",
        )

        req.add_argument(
            "-b",
            "--boxsize",
            type=int,
            required=True,
            help="Boxsize",
        )

        req.add_argument(
            "-o",
            "--out",
            required=True,
            help="Output path",
        )

    @staticmethod
    def coords2boundingbox(coords: np.array, boxsize: int) -> List[BoundBox]:
        boxes = []
        for row in coords:
            x = row[0] - boxsize/2
            y = row[1] - boxsize/2
            z = row[2]
            boxes.append(BoundBox(x=x, y=y, z=z, h=boxsize, w=boxsize))
        return boxes

    @staticmethod
    def read_coords(path) -> np.array:
        return np.atleast_2d(np.genfromtxt(path))

    def run(self, args):
        input_path = args.input
        output_path = args.out
        boxsize = args.boxsize

        os.makedirs(output_path, exist_ok=True)

        input_files = []
        if os.path.isfile(input_path):
            input_files.append(input_path)
        else:
            input_files = glob(os.path.join(input_path,"*.coords"))

        for file in input_files:
            coords = Coords2CBOX.read_coords(file)
            boxes = Coords2CBOX.coords2boundingbox(coords, boxsize=boxsize)
            filename = os.path.splitext(os.path.basename(file))[0]
            outpth = os.path.join(output_path,filename+".cbox")
            print(len(boxes))
            write_cbox_file(path=outpth,coordinates=boxes)





"""
In this tool we use the Model View Controller (MVC) design pattern.
The MCV can be considered an approach to distinguish between the data model, processing control and the user interface.
It neatly separates the graphical interface displayed to the user from the code that manages the user actions.

In this file we define the 'Boxmanager_view' which is the presentation of the model in a particular format.

It is a subclass of class 'Boxmanager_view_ABC' defines in Boxmanager_view_ABC.py (see the file for more docu)
    Boxmanager_view_ABC → interact with GUI framework and wrapper libs
    Boxmanager_view     →  Use ABC_view for changing its vars, which represent the model
"""


from cryoloBM import boxmanager_view_ABC,boxmanager_controller,boxmanager_toolbar, helper, constants
import numpy as np
from os import path

class Boxmanager_view(boxmanager_view_ABC.Boxmanager_view_ABC):

    def __init__(self,font):
        """
        :param font. Qfont type got at the start of the code in the boxmanager.py. It is the only rererence to PyQT
        """
        super(Boxmanager_view, self).__init__(font)
        self.controller = None

    def setup(self,controller:boxmanager_controller):
        """
        :param controller: Boxmanager_controller obj
        """
        self.setup_()
        self.controller = controller

    def reset_config(self) -> None:
        self.controller.reset_config()

    def close_boxmanager(self):
        self.controller.close_boxmanager()

    def picking_filament_changed(self) -> None:
        """
        It is connected to the combobox picking_filament_combox to select if picking particle or filament
        """
        # make it visible or not in function of the current picking type
        is_filament = self.picking_filament_combox.currentText() == constants.Picking_cb.FILAMENT.value

        # The picking in the 3D is not enable
        if not self.is_updating_params:
            type_visualization = constants.Display_visualization_cb.FILAMENT_2D.value if is_filament else constants.Display_visualization_cb.PARTICLE.value
            self.set_visualization_combobox(type_visualization=type_visualization)

        # set the 'has_filament' var of 'box_dictionary
        self.controller.set_has_filament(has_filament=is_filament)


        self.controller.model.picking_combobox_value = self.picking_filament_combox.currentText()
        self.controller.model.visualization = self.visualization_combobox.currentText()
        # update the visualization
        current_filename = self.controller.get_current_filename(with_extension=False)
        if self.controller.model.box_dictionary and not self.controller.model.box_dictionary.has_same_visualization(visualization=self.controller.model.visualization,
                                                                    f=current_filename,n=self.controller.model.index_tomo):
            self.controller.model.box_dictionary.update_visualization_dictionary(
                new_visualization=self.controller.model.visualization, f=current_filename, n=self.controller.model.index_tomo)

    def searchRange_label_changed(self)->None:
        self.controller.searchRange_label_changed()

    def searchRange_changed(self)->None:
        self.controller.searchRange_changed()

    def show_size_distribution(self)->None:
        """
        Function to show the size distribution
        """
        self.controller.show_size_distribution()

    def show_confidence_histogram(self)->None:
        """
        Function to to show the confidence histogram
        """
        self.controller.show_confidence_histogram()

    def open_SPA_folder(self, via_cli:bool = False)->None:
        """
        Function to read micrograph's images from a folder
        :param via_cli: True if we run from command line
        """
        self.controller.open_SPA_folder(via_cli=via_cli)

    def open_SPA_image(self, via_cli:bool = False)->None:
        """
        Function to read a single micrograph's image
        :param via_cli: True if we run from command line
        """
        self.controller.open_SPA_image(via_cli=via_cli)

    def open_tomo_folder(self, via_cli:bool = False)->None:
        """
        Function to read tomo's images from a folder
        :param via_cli: True if we run from command line
        """
        self.controller.open_tomo_folder(via_cli=via_cli)

    def open_tomo_image(self, via_cli:bool = False)->None:
        """
        Function to read a single tomo's image
        :param via_cli: True if we run from command line
        """
        self.controller.open_tomo_image(via_cli=via_cli)

    def import_box_files(self)->None:
        """
        Function to import box files from file
        """
        self.controller.import_box_files()

    def save_on_files(self)->None:
        """
        Function to to save the results on file. It write all the possible formats.
        It was write_all_type in the old version
        """
        self.controller.save_on_files()

    def apply_filter(self)->None:
        """
        It applies the lowpass filter present in filtering tab.
        """
        self.controller.apply_filter()

    def box_size_changed(self)->None:
        """
        It changes the value of the "box size" filling the 'box size' line in the visualization tab.
        It is triggered pressing enter (when the cursor is into the line) or button 'set'
        """
        self.controller.box_size_changed()

    def box_distance_filament_changed(self):
        """
        It changes the value of the "Box distance" filling the 'Box distance' line in the visualization tab.
        It is triggered pressing enter (when the cursor is into the line) or button 'set'
        """
        self.controller.box_distance_filament_changed()

    def visualization_changed(self)->None:
        """
        It changes the value of the "visualization_combobox" changing the 'visualization' combobox in the visualization tab.
        It is triggered when the combobox is changed
        """
        self.controller.visualization_changed()
        is_visible = self.get_visualization_cb_text() in [constants.Visualization_cb.CIRCLE_SEGMENTED.value,constants.Visualization_cb.RECT_FILAMENT_SEGMENTED.value]
        self.set_visibility_box_distance(is_visible=is_visible)

    def use_estimated_size_changed(self)->None:
        """
        It changes the value of the "box size" filling the 'box size' line in the visualization tab.
        It is usable only after loading data from cbox file. It gets the new "box size" from these files
        Pay attention: each particle will have its own estimated size
        It is triggered pressing enter (when the cursor is into the line) or button 'set'
        """
        self.controller.use_estimated_size_changed()

    def lower_size_thresh_changed(self)->None:
        self.controller.lower_size_thresh_changed()

    def lower_size_label_changed(self)->None:
        self.controller.lower_size_label_changed()

    def upper_size_thresh_changed(self)->None:
        self.controller.upper_size_thresh_changed()

    def upper_size_label_changed(self)->None:
        self.controller.upper_size_label_changed()

    def conf_thresh_changed(self)->None:
        self.controller.conf_thresh_changed()

    def conf_thresh_label_changed(self)->None:
        self.controller.conf_thresh_label_changed()

    def memory_changed(self)->None:
        self.controller.memory_changed()

    def memory_label_changed(self)->None:
        self.controller.memory_label_changed()

    def min_length_changed(self)->None:
        self.controller.min_length_changed()

    def min_length_label_changed(self)->None:
        self.controller.min_length_label_changed()

    def min_edge_weight_changed(self)->None:
        self.controller.min_edge_weight_changed()

    def min_edge_weight_label_changed(self)->None:
        self.controller.min_edge_weight_label_changed()

    def win_size_changed(self)->None:
        self.controller.win_size_changed()

    def win_size_label_changed(self)->None:
        self.controller.win_size_label_changed()

    def preview(self)->None:
        self.controller.preview()

    def trace(self)->None:
        self.controller.trace()


    # wrapper function set_first_image
    def create_plot_toolbar(self):
        return boxmanager_toolbar.Boxmanager_Toolbar( self.controller)

    def onclick(self, event)->None:
        """
        It is called when we click on the image
        :param event: the click event
        """
        self.controller.onclick(event=event)

    def myKeyPressEvent(self, event)->None:
        """
        It is called when we click on the image
        :param event: the click event
        """
        self.controller.myKeyPressEvent(event=event)

    def onrelease(self, event)->None:
        """
        It is called when we click on the image
        :param event: the click event
        """
        self.controller.onrelease(event=event)

    def onmove(self, event)->None:
        """
        It is called when we release the mouse button
        :param event: the click event
        """
        self.controller.onmove(event=event)

    def event_image_changed(self):
        """
        It is called when we change image (SPA or tomo cases) or slice (only tomo cases)
        """
        return self.controller.event_image_changed(self.tree.currentItem())

    def update_tree_boxsizes(self, update_current:bool =False, all_tomos:bool =False)->None:
        """
        Update the counter into the filename images tree
        """
        def update(filename:str)->None:
            has_filament = self.controller.model.picking_combobox_value == constants.Picking_cb.FILAMENT.value
            current_conf_thresh, upper_size_thresh, lower_size_thresh, current_num_boxes_thresh = self.controller.model.get_threshold_params(filename)
            res = [helper.check_if_should_be_visible(box=box, current_conf_thresh=current_conf_thresh, upper_size_thresh=upper_size_thresh, lower_size_thresh=lower_size_thresh,
                                                     num_boxes_thresh=current_num_boxes_thresh, is_filament=has_filament) for box in boxes]
            num_box_vis = int(np.sum(res))
            msg = "" if len(res) == 0 else "{0:> 4d}  / {1:> 4d}".format(num_box_vis, len(res))
            item.setText(1, msg)



        if self.controller.model.box_dictionary:
            # in case of imported filament via GUI i have to count the start_end filament
            start_vis = self.controller.model.box_dictionary.fil_start_end_vis
            self.controller.model.box_dictionary.fil_start_end_vis = True

            # the counter has to be set counting from the 2D particle dict
            no_3D = False
            no_tracing = False
            current_filename=self.controller.get_current_filename(with_extension=False)

            if update_current:
                item=self.tree.currentItem()
                boxes = self.controller.model.box_dictionary.get_sketches(f=current_filename,
                                                                     n=self.controller.model.index_tomo, is_3D = no_3D,  is_tracing = no_tracing)
                update(filename=current_filename)
            else:
                root = self.tree.invisibleRootItem().child(0)
                if not root:
                    return

                for i in range(root.childCount()):
                    self.run_in_same_thread()
                    item_root = root.child(i)
                    filename = path.splitext(item_root.text(0))[0]

                    if self.controller.model.type_case == constants.Type_case.TOMO_FOLDER.value:
                        fname = path.splitext(root.child(i).text(0))[0]
                        for chld in range(item_root.childCount()):
                            if fname != current_filename and not all_tomos:
                                # in this case we want to update only the current tomo
                                continue
                            item = item_root.child(chld)
                            boxes = self.controller.model.box_dictionary.get_sketches(f=fname, n=chld, is_3D = no_3D,  is_tracing = no_tracing)
                            update(filename=fname)
                        continue

                    if self.controller.model.type_case == constants.Type_case.SPA.value:
                        boxes = self.controller.model.box_dictionary.get_sketches(f=filename, n=None, is_3D = no_3D,  is_tracing = no_tracing)
                    else:
                        boxes = self.controller.model.box_dictionary.get_sketches(f=current_filename, n=int(filename), is_3D = no_3D,  is_tracing = no_tracing)
                    item = item_root
                    update(filename=current_filename)
            self.controller.model.box_dictionary.fil_start_end_vis = start_vis



    def update_all_global_counters(self)->None:
        """
        It updates the number of total particles in all the tomo, TOMO FOLDER case
        """
        # in case of imported filament via GUI i have to count the start_end filament
        start_vis = self.controller.model.box_dictionary.fil_start_end_vis
        self.controller.model.box_dictionary.fil_start_end_vis = True

        if self.controller.model.type_case == constants.Type_case.TOMO.value:
            self.update_global_counter(childid=None)
        else:
            for child_id in range(self.tree.invisibleRootItem().child(0).childCount()):
                self.update_global_counter(childid=child_id)
        self.controller.model.box_dictionary.fil_start_end_vis = start_vis

    def update_global_counter(self, childid:int = None)->None:
        """
        It updates the number of total particles in the current tomo
        :param childid: identifies the child id in the tomo folder case. It is used after reset->file or 'import' boxes
        """
        def detect_child()->int:
            """ detect the child id, which detect the current file. Only in tomo folder case"""
            for i in range(root.childCount()):
                if path.splitext(root.child(i).text(0))[0] == current_filename:
                    return i

        root = self.tree.invisibleRootItem().child(0)

        # it happen if you run from CLI without param
        if not root:
            return

        current_filename = path.splitext(root.child(childid).text(0))[0] if childid else self.controller.get_current_filename(with_extension=False)
        tot_child,tot_num_box_vis, tot_res, child_id = 0,0, 0, None

        # get the number of slice in the current tomo
        if self.controller.model.type_case == constants.Type_case.TOMO_FOLDER.value:
            child_id = detect_child() if childid is None else childid
            tot_child = root.child(child_id).childCount()
        elif self.controller.model.type_case == constants.Type_case.TOMO.value:
            tot_child = root.childCount()

        # count the number of boxes in the whole current tomo
        has_filament = self.controller.model.picking_combobox_value == constants.Picking_cb.FILAMENT.value
        current_conf_thresh, upper_size_thresh, lower_size_thresh, current_num_boxes_thresh = self.controller.model.get_threshold_params()

        if self.controller.model.type_case != constants.Type_case.SPA.value:
            for j in range(tot_child):
                boxes = self.controller.model.box_dictionary.get_sketches(f=current_filename, n=j, is_3D = False,  is_tracing = False)
                res = [helper.check_if_should_be_visible(box=box, current_conf_thresh=current_conf_thresh, upper_size_thresh=upper_size_thresh, lower_size_thresh=lower_size_thresh,
                                                         num_boxes_thresh=current_num_boxes_thresh, is_filament=has_filament) for box in boxes]
                tot_res += len(res)
                tot_num_box_vis += int(np.sum(res))
        else:
            for f in self.controller.model.box_dictionary.list_purefiles:
                boxes = self.controller.model.box_dictionary.get_sketches(f=f, n=None, is_3D=False,is_tracing=False)
                res = [helper.check_if_should_be_visible(box=box, current_conf_thresh=current_conf_thresh, upper_size_thresh=upper_size_thresh, lower_size_thresh=lower_size_thresh,
                                                         num_boxes_thresh=current_num_boxes_thresh, is_filament=has_filament) for box in boxes]
                tot_res += len(res)
                tot_num_box_vis += int(np.sum(res))

        # write the global counter on the tomo name line
        msg = "" if tot_res == 0 else "{0:> 4d}  / {1:> 4d}".format(tot_num_box_vis, tot_res)
        if self.controller.model.type_case == constants.Type_case.TOMO_FOLDER.value:
            root.child(child_id).setText(1, msg)
        else:
            root.setText(1, msg)

    def thresholding_tab_blurring(self, is_enable: bool, kind_of_cbox: bool,has_filament: bool, type_case: int) -> None:
        """
        After loading from file we have to unblur some options in thresholding tabs
        :param is_enable: If True enable SOME options. If False disable ALL the options
        :param kind_of_cbox: True if model.is_cbox or model.is_cbox_untracved is True
        :param has_filament: If True we work with filament
        :param type_case: constants.Type_case. value
        """
        self.thresholding_tab_blurring_(is_enable=is_enable, kind_of_cbox=kind_of_cbox, has_filament=has_filament, type_case=type_case)

        if not is_enable:
            return

        # In case of cbox files (not filament case), set the minimum and maximum
        if not has_filament and kind_of_cbox:
            self.controller.view.is_updating_params = True

            if type_case == constants.Type_case.SPA.value:
                max_size, min_size = self.controller.model.box_dictionary.get_upper_lower_thres()
            else:
                temp = self.controller.model.box_dictionary.get_upper_lower_thres()
                if type_case == constants.Type_case.TOMO_FOLDER.value:
                    for filename, values in temp.items():
                        self.controller.model.params[filename].upper_size_thresh = values["upper"]
                        self.controller.model.params[filename].lower_size_thresh = values["lower"]
                    pass
                current_fname = self.controller.get_current_filename(with_extension=False)
                max_size, min_size = temp[current_fname]["upper"], temp[current_fname]["lower"]

            self.update_low_up_thresh(max_size=max_size, min_size=min_size)

    def onresize(self, event)->None:
        """
        It is called when we resize the main window
        :param event: the click event
        """
        self.controller.onresize(event=event)

    def ondraw(self, event)->None:
        """
        It is called when we use the toolbar
        :param event: the click event
        """
        self.controller.ondraw(event=event)


"""
In this file the 'picked_filament' class is defined.
This class is used for saving all the data of a filament in the picking phase
"""

from cryolo import utils
import numpy as np
from cryoloBM import MySketch,helper_GUI,helper
from typing import Union,List

class picked_filament:
    """
    A picked filament
    In the rectangular visualization we lose the left half of the first square box (the one on the first point of the filament)
    """
    def __init__(self, box_size:int, is_tomo:bool=True, fil_id=None)->None:
        """
        :param box_size: size of the base of the rectangle
        :param is_tomo: True if the case is a tomo False if it is a SPA
        :param fil_id: id of the filament
        """
        self.id = fil_id

        self.boxsize = box_size
        self.begin_fil = list()        # event position obj ... the first click (got via on_click)
        self.end_fil = list()          # event position obj ... the las click (got via on_release or on_move)

        self.filament=utils.Filament(list())        # crYOLO filament object

        self.is_tomo=is_tomo

        self.angle = None # In rads
        self.offset = None          # offset to center the filament (i.e.: for default we pick via mouse on the left corner but we need to pick on the center of the filament)
        self.rect_vect = None       # versor of the rectangle
        self.orth_vect = None       # versor orthogonal to self.rect_vect

        # distance between the first and last sketch
        self.width = box_size
        self.height = box_size

    def create_from_rect(self,rect:MySketch.MySketch)->None:
        """
        create a picked filament object starting from a MySketche obj
        :param rect: MySketch obj
        """
        rad_angle = np.deg2rad(rect.get_angle())
        x, y = rect.get_xy(circle=False) # lower left corner of the box
        x += (self.boxsize / 2) * np.cos(rad_angle+np.pi/2) # center of box (x)
        y += (self.boxsize / 2) * np.sin(rad_angle+np.pi/2) # center of the box (y)

        x_fin = x + rect.get_width()* np.cos(rad_angle)
        y_fin = y + rect.get_width()* np.sin(rad_angle)
        x_in = x
        y_in = y
        self.begin_fil=[x_in,y_in]
        self.end_fil = [x_fin, y_fin]
        self.set_line_params()

    def set_first(self,event:helper_GUI.plt_qtbackend)->None:
        """
        It sets the first element
        :param event: click event via 'onclick
        """
        self.begin_fil = [event.xdata,event.ydata]
        self.end_fil =  [event.xdata,event.ydata]

    def append_sketch(self,sketch:MySketch.MySketch)->None:
        """
        Append sketch to le list
        :param sketch: MySketch obj
        """
        self.filament.add_box(sketch)

    def reset_sketches(self)->None:
        self.filament.boxes = list()


    def set_line_params(self)->None:
        """
        Calculate params of the line got from  'begin_fil' and 'end_fil'
        """
        self.rect_vect = np.array(self.end_fil) - np.array(self.begin_fil)
        self.width = np.linalg.norm(self.rect_vect)
        self.angle = np.rad2deg(np.arctan2(self.rect_vect[1], self.rect_vect[0]))

        # calculate offset to filament start. this has the effect that filaments can picked centerd.
        self.orth_vect = np.array(self.rect_vect)/self.width # normalized vector
        help_value = self.orth_vect[0]
        self.orth_vect[0] = self.orth_vect[1]
        self.orth_vect[1] = -1*help_value  #now it orthogonal to rect_vect
        self.offset = self.orth_vect #* self.height/2 # offset for drawing.

    def get_R(self)->np.array:
        """
        Returns the rotation matrix of the rectangle as np array
        """
        theta = -np.radians(self.angle)
        c, s = np.cos(theta), np.sin(theta)
        return np.array(((c, -s), (s, c)))


    def get_coordinate_rect(self)->np.array:
        """
        returns the real coordinate of the rectangles as np array
        -) each row is a point x,y.
        -) first row =bottom left corner
        -) second row = bottom right corner
        -) third row = top left corner
        -) fourth row = top right corner
        """
        norm_v_ort = self.orth_vect/np.linalg.norm(self.orth_vect)
        scale_v_ort = norm_v_ort*self.height

        corner_0 = np.array(self.begin_fil)-np.array(self.offset * self.height/2)
        corner_1 = corner_0 + scale_v_ort
        corner_2 = corner_0 + self.rect_vect
        corner_3 = corner_2 + scale_v_ort
        return np.vstack((corner_0, corner_1, corner_2, corner_3))


    def  get_rotated_coordinate_rect(self)->np.array:
        """
        Returns the coordinate of the rect on the x axis
        """
        return self.get_R()@self.get_coordinate_rect().T

    def fill_and_get_sketches(self, box_distance:int=1, as_bbox:bool=False, z:int=None, color ='r')->Union[List[utils.BoundBox],List[MySketch.MySketch]]:
        """
        It is used for saving data on file
        Create a square box every 'box_distance' point on the central line of the filament
        :param box_distance: sampling step
        :param as_bbox: convert the sketches in cryolo.bbox
        :param z:   in case of filament in a tomo I need to know the slice
        :param color: edgecolor
        """
        self.fill_sketches(box_distance=box_distance,as_bbox=as_bbox, z=z, color=color)
        return self.get_sketches()

    def fill_sketches(self, box_distance:int = 1, as_bbox:bool = False, z:int=None, color ='r')->None:
        """
        It is used for saving data on file or fill the box_dictionary.particle_dictionary dict for speed up the realtime visualization
        Create a square box every 'box_distance' point on the central line of the filament
        :param box_distance: sampling step
        :param as_bbox: convert the sketches in cryolo.bbox
        :param z:   in case of filament in a tomo I need to know the slice
        :param color: edgecolor
        """
        self.set_line_params()
        self.reset_sketches()
        nboxes = int(self.width//box_distance)

        begin_sketch = MySketch.MySketch(xy=(self.begin_fil[0] - self.height / 2 , self.begin_fil[1] - self.height / 2 ), width=self.boxsize, height=self.boxsize,
                                         is_tomo=self.is_tomo, angle=self.angle, est_size=self.boxsize, confidence=1,
                                         only_3D_visualization=False, num_boxes=1, meta=None, z=z,
                                         linewidth=1, edgecolor=color, facecolor="none")
        end_sketch = MySketch.MySketch(xy=(self.end_fil[0] - self.height / 2 , self.end_fil[1] - self.height / 2 ), width=self.boxsize, height=self.boxsize,
                                       is_tomo=self.is_tomo, angle=self.angle, est_size=self.boxsize, confidence=1,
                                       only_3D_visualization=False, num_boxes=1, meta=None, z=z,
                                       linewidth=1, edgecolor=color, facecolor="none")
        self.filament.boxes = utils.getEquidistantBoxes(begin_sketch.get_as_BBox(confidence=1, z=z),end_sketch.get_as_BBox(confidence=1, z=z),nboxes)
        if as_bbox is False:
            self.filament.boxes = helper.convert_list_bbox_to_sketch(self.filament.boxes,out_dict=False, fil_id =self.id,col=color)

    def get_sketches(self,box_distance:int = 1, as_bbox:bool = False, z:int=None)->Union[List[utils.BoundBox],List[MySketch.MySketch]]:
        """
        It is used for saving data on file or fill the box_dictionary.particle_dictionary dict for speed up the realtime visualization
        Create a square box every 'box_distance' point on the central line of the filament
        :param box_distance: sampling step
        :param as_bbox: convert the sketches in cryolo.bbox
        :param z:   in case of filament in a tomo I need to know the slice
        """
        if not self.filament.boxes:
            return self.fill_and_get_sketches(box_distance=box_distance,as_bbox=as_bbox,z=z)
        return self.filament.boxes


    def get_rect_sketch(self,color='r',sketch_id=None)->MySketch.MySketch:
        """
        :param color: edgecolor
        :param sketch_id: id, used only in resize step
        """
        self.set_line_params()

        # for centering on x i have to calculate, geometrically, it. It does not deserve calulate it
        return  MySketch.MySketch(xy=(self.begin_fil[0] + self.offset[0] * self.height/2, self.begin_fil[1] + self.offset[1]* self.height/2), width=self.width, height=self.height,
                                  is_tomo=self.is_tomo, angle=self.angle, est_size=self.boxsize, confidence=1,
                                  only_3D_visualization=False, num_boxes=1, meta=None, z=None,id=sketch_id,
                                  linewidth=1, edgecolor=color, facecolor="none")



import unittest
from cryoloBM.MySketch import MyRectangle

class MyTestCase(unittest.TestCase):
    def test_resize_10_to_20(self):
        rect = MyRectangle(
            confidence = None,
            est_size = None,
            is_tomo = False,
            xy = (10,10),
            width = 100,
            height = 10,
            angle = 0,
        )
        rect.set_height(20)
        self.assertEqual(rect.xy[0], 10)
        self.assertEqual(rect.xy[1], 5)

    def test_resize_10_to_5(self):
        rect = MyRectangle(
            confidence = None,
            est_size = None,
            is_tomo = False,
            xy = (10,10),
            width = 100,
            height = 10,
            angle = 0,
        )
        rect.set_height(5)
        self.assertEqual(rect.xy[0], 10)
        self.assertEqual(rect.xy[1], 12.5)

    def test_resize_10_to_20_90degree(self):
        rect = MyRectangle(
            confidence = None,
            est_size = None,
            is_tomo = False,
            xy = (10,10),
            width = 100,
            height = 10,
            angle = 90,
        )
        rect.set_height(20)
        self.assertAlmostEqual(rect.xy[0], 5)
        self.assertAlmostEqual(rect.xy[1], 10)

    '''
    def test_resize_10_to_20_45degree(self):
        rect = MyRectangle(
            confidence = None,
            est_size = None,
            is_tomo = False,
            xy = (10,10),
            width = 100,
            height = 10,
            angle = 45,
        )
        rect.set_height(20)
        print(rect.xy)
        self.assertAlmostEqual(rect.xy[0], 5)
        self.assertAlmostEqual(rect.xy[1], 10)
    '''


if __name__ == '__main__':
    unittest.main()



import numpy as np
from pyStarDB import sp_pystardb as star
import os
import glob
from cryoloBM.bmtool import BMTool
from argparse import ArgumentParser
import argparse

class Priors2StarTool(BMTool):

    def get_command_name(self) -> str:
        return "priors2star"

    def create_parser(self, parentparser : ArgumentParser) -> ArgumentParser:

        parser_priors2star = parentparser.add_parser(
            self.get_command_name(),
            help="Add filament prior information to star file.",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )

        priors2star_required_group = parser_priors2star.add_argument_group(
            "Required arguments",
            "Add filament prior information to star",
        )

        priors2star_required_group.add_argument(
            "-i",
            "--input",
            required=True,
            help="Path to particles.star file.",
        )
        priors2star_required_group.add_argument(
            "-fi",
            "--fidinput",
            required=True,
            help="Input folder or file with *_fid.coords files from crYOLO .",
        )

        priors2star_required_group.add_argument(
            "-o",
            "--output",
            required=True,
            help="Output folder where to write the augmented star files..",
        )

        return parser_priors2star

    def run(self, args):


        if os.path.isfile(args.input):
            star_file = args.input
        else:
            raise ValueError("Can't find input star file.")

        if os.path.isfile(args.fidinput):
            fid_files = [args.fidinput]
        else:
            path = os.path.join(os.path.abspath(args.fidinput), "*_fid.coords")
            fid_files = glob.glob(path)

        # Find star fid paris
        outname = os.path.splitext(os.path.basename(star_file))[0] + "_with_prior.star"
        add_prior_to_star(
            in_star=star_file,
            coords_fid_paths=fid_files,
            output_star=os.path.join(args.output, outname),
        )

def match(tomofiles,fid_files):
    '''
    This functions gives the indices
    :param star_files:
    :param fid_files:
    :return: One list for every fid_file containing the relevant indices in tomofiles
    '''

    fid_index_list = []
    fid_file_basenames = [os.path.basename(file) for file in fid_files]

    for fid_file in fid_file_basenames:
        indicies = [row_index for row_index, tomopth in enumerate(tomofiles) if os.path.splitext(os.path.basename(tomopth))[0] in fid_file]
        fid_index_list.append(indicies)
    return fid_index_list

import pandas as pd
def get_column_index(columnname: str, star : pd.DataFrame) -> int:

    if columnname not in star:
        star[columnname] = 0
    return star.columns.get_loc(columnname)

def add_prior_to_star(in_star, coords_fid_paths, output_star):
    '''

    :param in_star:
    :param coords_fid_path:
    :param ouput_star:
    :return:
    '''
    import copy
    from cryoloBM_tools import coords2warp

    os.makedirs(os.path.dirname(output_star), exist_ok=True)
    sfile = star.StarFile(in_star)
    relion_dataframe = sfile['']

    tomo_names = relion_dataframe['_rlnMicrographName']
    fid_index_lists = match(tomofiles=tomo_names,fid_files=coords_fid_paths)
    relion_dataframe_with_priors = copy.deepcopy(relion_dataframe)

    tubeindex = get_column_index('_rlnHelicalTubeID', relion_dataframe_with_priors)
    tiltindex = get_column_index('_rlnAngleTiltPrior', relion_dataframe_with_priors)
    psiindex = get_column_index('_rlnAnglePsiPrior', relion_dataframe_with_priors)
    flipindex = get_column_index('_rlnAnglePsiFlipRatio', relion_dataframe_with_priors)



    for i, fid_file in enumerate(coords_fid_paths):
        coords = np.atleast_2d(np.genfromtxt(fid_file))
        fid_indices = fid_index_lists[i]


        relion_dataframe_with_priors.iloc[fid_indices, tubeindex] = coords[:,3]
        relion_dataframe_with_priors.iloc[fid_indices,tiltindex] = 0
        relion_dataframe_with_priors.iloc[fid_indices,psiindex] = 0
        relion_dataframe_with_priors.iloc[fid_indices,flipindex] = 0
        npdata = coords2warp.add_prior_information(relion_dataframe_with_priors.iloc[fid_indices,:])
        relion_dataframe_with_priors.iloc[fid_indices, :] = npdata



    sfile.update('',relion_dataframe_with_priors, True)
    sfile.write_star_file(output_star)



from abc import ABC, abstractmethod
from argparse import ArgumentParser

class BMTool(ABC):

    @abstractmethod
    def get_command_name(self) -> str:
        '''
        :return: Command name
        '''
        pass

    @abstractmethod
    def create_parser(self, parentparser : ArgumentParser) -> ArgumentParser:
        '''
        :param parentparser: ArgumentPaser where the subparser for this tool needs to be added.
        :return: Argument parser that was added to the parentparser
        '''
        pass

    @abstractmethod
    def run(self, args):
        pass


"""
The following functions manage the writing data on file option
"""

from os import path,makedirs, remove as os_remove
from cryoloBM import helper_GUI,constants,helper,helper_filament
from cryolo import CoordsIO
from cryolo.utils import BoundBox,Filament

from cryoloBM import boxmanager_controller,boxmanager_view,MySketch,box_dictionary # for type hinting purpose
from typing import Union,List,Dict,Tuple,Callable

def save_on_files(controller:boxmanager_controller)->None:
    """
    Save on file the results
    :param controller: Boxmanager_controller obj
    """
    selected_slices=get_selected_slice(controller=controller)
    box_dir = get_box_dir_path(view=controller.view)

    if not box_dir:
        return

    # i do not want to change the GUI var, it is used for saving purpose
    box_distance_filament = None
    if controller.model.has_filament:
        box_distance_filament, result =controller.view.get_inter_box_distance(value=controller.model.box_distance_filament)
        if not result:
            return
        elif box_distance_filament < 1:
            controller.view.err_message( msg="Invalid Box distance value. It has to be higher than 0")
            return
        controller.model.box_distance_filament = box_distance_filament

    if controller.model.is_cbox_untraced and not controller.model.has_filament: # not self.has_filament because in this version we cannot trace in filament case
        write_all_type_optimization(controller=controller,box_dir=box_dir, selected_slices=selected_slices)
    else:
        fil_start_end_vis = controller.model.box_dictionary.fil_start_end_vis
        controller.model.box_dictionary.fil_start_end_vis= False
        write_all_type_classic(controller=controller,box_dir=box_dir, selected_slices=selected_slices)
        controller.model.box_dictionary.fil_start_end_vis = fil_start_end_vis

    controller.model.unsaved_changes = False

def write_all_type_classic(controller:boxmanager_controller, box_dir:str, selected_slices:Union[List[str],Dict[str,str]])->None:
    """
    It saves on file the results after a classic picking ( no optimizing result case)
    :param controller: Boxmanager_controller obj
    :param box_dir: main directory where save the results
    :param selected_slices: list or dict ( in case of folder of tomo) of slices to save
    """

    if controller.model.has_filament:
        #set the box_distance_filament
        controller.model.box_dictionary.change_box_distance_all(new_box_distance=controller.model.box_distance_filament)
        # set the box size
        controller.model.box_dictionary.change_box_size_all(new_box_size=controller.model.boxsize, is_saving_on_file=True)

    box_dictionary_to_save, ignored_slice, empty_slice = prepare_output(controller= controller, slices=selected_slices,is_tracing=False)


    is_tomo = controller.model.type_case != constants.Type_case.SPA.value
    for file_type in ["STAR", "CBOX", "EMAN"]:
        is_cbox = file_type == "CBOX"

        # the picked filament will be saved only in cbox format
        if controller.model.has_filament and not is_cbox :
            continue

        d = path.join(box_dir, file_type)
        file_ext, write_coords_ = prepare_vars_for_writing(controller = controller, box_dir=box_dir, file_type=file_type)

        pd = controller.view.create_progress_dialog(title="Write box files to " + d)
        pd.show()

        if is_tomo:
            if is_cbox:
                if controller.model.has_filament:
                    d+=constants.Writer_filament_suffix.CBOX_FILAMENT_SUFFIX.value
                    box_dictionary_to_save=create_box_dict_of_filament_obj(controller.model.box_dictionary,controller.model.box_distance_filament)
                else:
                    # we want to save ANYWAY all the selected boxes
                    box_dictionary_to_save = controller.model.box_dictionary.particle_dictionary
                write_coordinates_tomo(controller=controller,
                                       box_dict_to_save=box_dictionary_to_save,
                                       pd=pd, empty_slice=empty_slice, ignored_slice=ignored_slice,
                                       is_cbox=is_cbox, write_coords_=write_coords_,
                                       file_ext=file_ext, box_dir=d,has_filament=controller.model.has_filament)
            pd.close()
            continue

        # create an empty file name for the checked micrograph file
        save_empty_file_SPA(controller=controller,file_ext=file_ext, d=d, empty_slice=empty_slice)

        if is_cbox:
            if controller.model.has_filament:
                d+=constants.Writer_filament_suffix.CBOX_FILAMENT_SUFFIX.value
                box_dictionary_to_save = create_box_dict_of_filament_obj(controller.model.box_dictionary,
                                                                         controller.model.box_distance_filament)
            write_coordinates_micrograph(controller=controller, pd=pd, box_dict_to_save=box_dictionary_to_save,
                                         box_dir=d, file_ext=file_ext, is_cbox=is_cbox, write_coords_=write_coords_,has_filament=controller.model.has_filament)

        elif file_type == "EMAN" and controller.model.has_filament:
            for write_coords in write_coords_:
                suffix = constants.Writer_filament_suffix.HELICON_SUFFIX.value if "write_eman1_helicon" in str(
                    write_coords) else constants.Writer_filament_suffix.EMAN_STAR_FILAMENT_SUFFIX.value
                box_dir = d + suffix
                write_coordinates_micrograph(controller=controller, pd=pd, box_dict_to_save=box_dictionary_to_save,
                                             box_dir=box_dir, file_ext=file_ext, is_cbox=is_cbox, write_coords_=write_coords_,has_filament=False)


        else:
            if controller.model.has_filament:
                d += constants.Writer_filament_suffix.EMAN_STAR_FILAMENT_SUFFIX.value
            write_coordinates_micrograph(controller=controller, pd=pd, box_dict_to_save=box_dictionary_to_save,
                                         box_dir=d, file_ext=file_ext, is_cbox=is_cbox, write_coords_=write_coords_,has_filament=False)

        pd.close()

    del box_dictionary_to_save

def write_all_type_optimization(controller:boxmanager_controller, box_dir:str, selected_slices:Union[List[str],Dict[str,str]]):
    """
    It saves on file the results after the optimization ( cbox_untraced case)
    We save on file the result of the tracing
    :param controller: Boxmanager_controller obj
    :param box_dir: main directory where save the results
    :param selected_slices: list or dict ( in case of folder of tomo) of slices to save
    """

    for file_type in ["STAR", "CBOX_3D", "EMAN_3D", "CBOX_UNTRACED", "COORDS"]:

        is_cbox = "CBOX" in file_type or "EMAN_3D" in file_type       # because in EMAN_3D output we want to have the same as in CBOX_3D


        d = path.join(box_dir, file_type)

        file_ext, write_coords_ = prepare_vars_for_writing(controller=controller, box_dir=box_dir, file_type=file_type)

        if file_type == "COORDS":
            write_coords_ = CoordsIO. write_coords_file

        pd = controller.view.create_progress_dialog(title="Write box files to " + d)
        pd.show()

        if file_type == "CBOX_UNTRACED":
            box_dictionary_to_save, ignored_slice, empty_slice = prepare_output(controller= controller, slices=selected_slices,is_tracing=True)

            # at 19.02.21 the picking filament is not enabled
            # after the tracing we save into 'CBOX_UNTRACED' the input value gave to 'grouping3d.do_tracing' called int 'helper_tab_tracing.trace_values'
            # hence the model.box_dictionary.particle_dictionary
            write_coordinates_tomo(controller=controller,
                                   box_dict_to_save=controller.model.box_dictionary.particle_dictionary,
                                   pd=pd, empty_slice=empty_slice, ignored_slice=ignored_slice,
                                   is_cbox=is_cbox, write_coords_=write_coords_,
                                   file_ext=file_ext, box_dir=d,has_filament=False)
            pd.close()
            continue


        # we want to save ANYWAY all the selected boxes
        write_coordinates_tomo(controller=controller,
                               box_dict_to_save=controller.model.box_dictionary.box_dict_traced,
                               pd=pd, empty_slice=[], ignored_slice=[],
                               is_cbox=is_cbox, write_coords_=write_coords_,
                               file_ext=file_ext, box_dir=d,has_filament=False)

        pd.close()
    controller.view.warning_retrain()
    return

def get_selected_slice(controller:boxmanager_controller)->Union[List[str],Dict[str,str]]:
    """
    Return dict (in TOMO_FOLDER case otherwise list) of all the checked micrographs/slices
    :param controller: Boxmanager_controller obj
    :return list/dict of micrograph's name/#slices
    """
    selected_slices = dict() if controller.model.type_case == constants.Type_case.TOMO_FOLDER.value else list()
    for root_index in range(controller.view.tree.invisibleRootItem().childCount()):
        root_element = controller.view.tree.invisibleRootItem().child(root_index)  # can be tomogram or a folder
        for child_index in range(root_element.childCount()):
            if controller.model.type_case == constants.Type_case.TOMO_FOLDER.value :
                selected_slice_child = list()
                for child_child_index in range(root_element.child(child_index).childCount()):
                    if root_element.child(child_index).child(child_child_index).checkState(0) == controller.view.checkbox_state_value(checked=True):
                        selected_slice_child.append(str(child_child_index))     # to be able to generalize 'prepare_output's code
                selected_slices.update({path.splitext(path.basename(root_element.child(child_index).text(0)))[
                                           0]: selected_slice_child})
            elif root_element.child(child_index).checkState(0) == controller.view.checkbox_state_value(checked=True):
                selected_slices.append(path.splitext(path.basename(root_element.child(child_index).text(0)))[0])

    # In case of single tomo the output has to be a dict as in the folder tomo case
    if controller.model.type_case == constants.Type_case.TOMO.value:
        selected_slices = {controller.get_current_filename(with_extension=False): selected_slices}
    return selected_slices


def get_box_dir_path(view:boxmanager_view)->Union[str,None]:
    """
    Ask the dir where to save and return it
    :param view: Boxmanager_view obj
    :return box dir path
    """
    box_dir = view.qt_esixting_folder(msg="Select Box Directory")

    # Remove untitled from path if untitled not exists
    if box_dir and box_dir[-8] == "untitled" and path.isdir(box_dir):
        return box_dir[:-8]

    if not box_dir:
        return None

    return box_dir


def prepare_output(controller: boxmanager_controller, slices: List[str], is_tracing: bool) -> Tuple[
                                Union[Dict[str, List[MySketch.MySketch]], Dict[str, Dict[int, List[MySketch.MySketch]]]],
                                Union[List[str], Dict[str, List[str]]],
                                Union[List[str], Dict[str, List[str]]]]:
    """
    Select the data for writing on  file
    :param controller: Boxmanager_controller obj
    :param slices: list of slices to save
    :param is_tracing: if True it search into the tracing dict (box_dict_traced) otherwise (particle_dictionary)
    :return
            box_dict: box_dict format with slice containing boxes in 'self.box_dictionary' format. (dict of dict in case 3D folder)
            slice_ignored: list of slice with boxes but their checkbox are unchecked  (dict of list in case 3D folder)
            slice_empty: list of slice without boxes but their checkbox are checked (dict of list in case 3D folder)
    """
    b_dict = controller.model.box_dictionary.box_dict_traced if is_tracing else controller.model.box_dictionary.particle_dictionary
    box_dict = dict()
    is_tomo = controller.model.type_case != constants.Type_case.SPA.value
    slice_ignored = dict() if is_tomo else list()
    slice_empty = dict() if is_tomo else list()

    if controller.model.type_case == constants.Type_case.SPA.value:
        for f in controller.model.box_dictionary.list_purefiles:
            sketches = controller.model.box_dictionary.get_sketches(f=f, n=None, is_3D=False,is_tracing=is_tracing)
            empty = len(sketches) == 0
            in_slices = f in slices
            if empty and in_slices:
                slice_empty.append(f)
            elif not empty and not in_slices:
                slice_ignored.append(f)
            elif not empty and in_slices:
                box_dict.update({f: sketches})
    else:
        for f, d in b_dict.items():
            box_dict.update({f: dict()})
            slice_ignored.update({f:list()})
            slice_empty.update({f: list()})
            for n,boxes in d.items():
                in_slices = str(n) in slices[f]
                empty= len(boxes) == 0
                if empty and in_slices:
                    slice_empty[f].append(n)
                elif not empty and not in_slices:
                    slice_ignored[f].append(n)
                elif not empty and in_slices:
                    box_dict[f].update({n: boxes})

    return box_dict, slice_ignored, slice_empty

def prepare_vars_for_writing(controller:boxmanager_controller, box_dir:str, file_type:str)->Tuple[str,Union[Callable,List[Callable]]]:
    """
    Function to detect which cryolo function use for saving data on file and with which suffix.
    It create all the output folder !!
    :param controller: Boxmanager_controller obj
    :param box_dir: main directory where save the results
    :param file_type: filetype ("STAR", "CBOX" or "EMAN")
    """
    file_ext = ""
    write_coords_= None
    if controller.model.has_filament:
        if file_type == "EMAN":
            file_ext = ".box"
            write_coords_ = [CoordsIO.write_eman1_filament_start_end, CoordsIO.write_eman1_helicon]
            makedirs(path.join(box_dir, file_type + constants.Writer_filament_suffix.HELICON_SUFFIX.value ), exist_ok=True)
            makedirs(path.join(box_dir, file_type + constants.Writer_filament_suffix.EMAN_STAR_FILAMENT_SUFFIX.value), exist_ok=True)
        elif file_type == "STAR":
            file_ext = ".star"
            write_coords_ = CoordsIO.write_star_filemant_file
            makedirs(path.join(box_dir, file_type + constants.Writer_filament_suffix.EMAN_STAR_FILAMENT_SUFFIX.value), exist_ok=True)
        elif file_type == "CBOX":
            file_ext = ".cbox"
            write_coords_ = CoordsIO.write_cbox_file
            makedirs(path.join(box_dir, file_type + constants.Writer_filament_suffix.CBOX_FILAMENT_SUFFIX.value), exist_ok=True)
        return file_ext, write_coords_

    is_spa = controller.model.type_case == constants.Type_case.SPA.value
    d = path.join(box_dir, file_type)
    makedirs(d, exist_ok=True)
    file_ext = ".box"
    write_coords_ = CoordsIO.write_eman1_boxfile if is_spa  else  CoordsIO.write_eman_boxfile3d

    if file_type == "STAR":
        file_ext = ".star"
        write_coords_ = CoordsIO.write_star_file
    elif file_type in ["CBOX", "CBOX_3D", "EMAN_3D", "CBOX_UNTRACED"]:
        file_ext = ".cbox"
        write_coords_ = CoordsIO.write_cbox_file
    return file_ext, write_coords_


def save_empty_file_SPA(controller:boxmanager_controller, file_ext:str, d:str, empty_slice:List[str])->None:
    """
    Create the empty files in SPA case
    """
    if not controller.model.has_filament:
        for f in empty_slice:
            with open(path.join(d, f + file_ext), "w"):
                pass
        return

    directories = list()
    if file_ext == ".cbox":
        directories = [d + constants.Writer_filament_suffix.CBOX_FILAMENT_SUFFIX.value]
    elif file_ext == ".star":
        directories = [d + constants.Writer_filament_suffix.EMAN_STAR_FILAMENT_SUFFIX.value]
    elif file_ext == ".box":
        directories = [d + constants.Writer_filament_suffix.HELICON_SUFFIX.value,d + constants.Writer_filament_suffix.EMAN_STAR_FILAMENT_SUFFIX.value]

    for d in directories:
        for f in empty_slice:
            with open(path.join(d, f + file_ext), "w"):
                pass


def write_coordinates_tomo(controller: boxmanager_controller,
                           box_dict_to_save: Dict[str, Dict[int, List[MySketch.MySketch]]],
                           pd: helper_GUI.QtG.QProgressDialog,
                           empty_slice: Union[Dict[str, List[str]],list],
                           ignored_slice: Union[Dict[str, List[str]],list],
                           is_cbox: bool,
                           write_coords_: Callable,
                           file_ext: str,
                           box_dir: str, has_filament:bool) -> None:
    num_writtin_part = 0
    for f in box_dict_to_save.keys():
        counter = 0
        overwrite = True
        boxes = []
        for n, rectangles in box_dict_to_save[f].items():

            if pd.wasCanceled():
                break
            else:
                pd.show()
                val = int((counter + 1) * 100 / len(box_dict_to_save[f].items()))
                pd.setValue(val)
            controller.view.run_in_same_thread()

            box_filename = f + file_ext
            box_file_path = path.join(box_dir, box_filename)

            if overwrite is True:
                overwrite = False
                if path.isfile(box_file_path):
                    os_remove(box_file_path)

            if has_filament:
                num_writtin_part = num_writtin_part + len(rectangles)
                boxes += rectangles
            else:
                if is_cbox:
                    rectangles = [ box for box in rectangles if
                        helper.check_if_should_be_visible(box=box, current_conf_thresh=controller.model.current_conf_thresh,
                                                          upper_size_thresh=controller.model.upper_size_thresh,
                                                          lower_size_thresh=controller.model.lower_size_thresh,
                                                          num_boxes_thresh=controller.model.current_num_boxes_thresh,
                                                          is_filament=False)
                    ]

                real_sketches = [rect.getSketch(circle=False) for rect in rectangles]
                if controller.model.use_estimated_size:
                    for sketch in real_sketches:
                        sketch.resize(controller.model.boxsize)
                num_writtin_part += len(real_sketches)

                for rect_index, rect in enumerate(real_sketches):
                    confidence = 1
                    if rectangles[rect_index].get_confidence() is not None:
                        confidence = rectangles[rect_index].get_confidence()
                    x_lowerleft = int(rect.get_x())
                    y_lowerleft = int(rect.get_y())
                    boxize = int(rect.get_width())
                    box = BoundBox(x=x_lowerleft, y=y_lowerleft, w=boxize, h=boxize, c=confidence, z=n,depth=1)
                    boxes.append(box)

            counter = counter + 1

        if is_cbox:
            # there is a case where i pass to this function empty_slice=ignored_slice=[]
            empty = empty_slice[f] if empty_slice else list()
            ignored = ignored_slice[f] if ignored_slice else list()
            if len(empty)==0 and len(boxes)==0:
                continue
            write_coords_(box_file_path, boxes, empty, ignored)
        else:
            write_coords_(box_file_path, boxes)


    if controller.model.has_filament:
        print(f"{controller.model.box_dictionary.get_tot_filament(tomo_name=None)} filaments written in '{file_ext}' file.")
    else:
        print(f"{num_writtin_part} particles written in '{file_ext}' file.")

def write_coordinates_micrograph(controller:boxmanager_controller, pd:helper_GUI.QtG.QProgressDialog, box_dict_to_save:Dict[str, List[MySketch.MySketch]], box_dir:str, file_ext:str, is_cbox:bool, write_coords_:Callable, has_filament:bool)->None:
    # only in a 3D tomo image we have to delete an existing file before starting to write the new one.
    # Basically because we write in append mode on a file the selected particles slice after slice
    num_writtin_part = 0
    counter = 0
    slice_index = None
    boxes = list()
    for filename, rectangles in box_dict_to_save.items():

        if pd.wasCanceled():
            break
        else:
            pd.show()
            val = int((counter + 1) * 100 / len(box_dict_to_save.items()))
            pd.setValue(val)
        controller.view.run_in_same_thread()

        box_filename = filename + file_ext
        box_file_path = path.join(box_dir, box_filename)

        if has_filament:
            num_writtin_part = num_writtin_part + len(rectangles)
            boxes += rectangles
        else:
            if is_cbox:
                rectangles = [
                    box for box in rectangles if
                    helper.check_if_should_be_visible(box=box, current_conf_thresh=controller.model.current_conf_thresh,
                                               upper_size_thresh=controller.model.upper_size_thresh,
                                               lower_size_thresh=controller.model.lower_size_thresh,
                                               num_boxes_thresh=controller.model.current_num_boxes_thresh,
                                               is_filament=False)
                ]
                # if self.use_estimated_size_checkbox.isChecked():

            real_sketches = [rect.getSketch(circle=False) for rect in rectangles]
             # todo: why i m doing the resize only for SPA? Anyway the code works properly
            if controller.model.use_estimated_size:
                for sketch in real_sketches:
                    sketch.resize(controller.model.boxsize)

            num_writtin_part += len(real_sketches)

            for rect_index, rect in enumerate(real_sketches):
                x_lowerleft = int(rect.get_x())
                y_lowerleft = int(rect.get_y())
                boxize = int(rect.get_width())
                confidence = 1
                if rectangles[rect_index].get_confidence() is not None:
                    confidence = rectangles[rect_index].get_confidence()
                box = BoundBox(x=x_lowerleft, y=y_lowerleft, w=boxize, h=boxize, z=slice_index, c=confidence, depth=1)
                box.meta.update({'num_boxes':rectangles[rect_index].num_boxes})
                boxes.append(box)

        counter += 1
        if boxes:
            if "write_eman1_filament_start_end" in str(write_coords_):
                write_coords_(boxes,box_file_path)
            elif "write_eman1_helicon" in str(write_coords_):
                write_coords_(boxes, box_file_path,filename)
            else:
                write_coords_(box_file_path, boxes)
        boxes = list()

    if controller.model.has_filament:
        print(f"{controller.model.box_dictionary.get_tot_filament(tomo_name=None)} filaments written in '{file_ext}' file.")
    else:
        print(f"{num_writtin_part} particles written in '{file_ext}' file.")


def create_box_dict_of_filament_obj(box_dict:box_dictionary.Box_dictionary,box_distance_filament_picking:int ):
    """
    Return the box_dictionary for writing on file.
    It is used basically for creating a list of Filament obj from a list of  filament rect box
    :param box_dict
    :param box_distance_filament_picking: distance between boxes in case of filament
    """
    def fill(fil_segmented):
        """
        fill the dict
        """
        l=list()
        listed_sketches = dict()
        if fil_segmented:
            for box in fil_segmented:
                if box.id in listed_sketches:
                    listed_sketches[box.id].append(box.get_xy(circle=False))
                else:
                    listed_sketches.update({box.id: [box.get_xy(circle=False)]})

            boxsize = fil_segmented[0].get_height() if fil_segmented else None

            for fil_id, xy in listed_sketches.items():
                fil = helper_filament.picked_filament(box_size=boxsize, is_tomo=is_tomo, fil_id=fil_id)
                fil.begin_fil = listed_sketches[fil_id][0][0] + boxsize / 2, listed_sketches[fil_id][0][1] + boxsize / 2
                fil.end_fil = listed_sketches[fil_id][-1][0] + boxsize / 2, listed_sketches[fil_id][-1][1] + boxsize / 2
                l.append(Filament(fil.fill_and_get_sketches(box_distance=box_distance_filament_picking, as_bbox=True, z=z)))
        return l

    # set temporarily to True for getting via 'get_sketches' the start_end rect
    starting_vis = box_dict.fil_start_end_vis
    box_dict.fil_start_end_vis = False

    is_tomo = box_dict.type_case != constants.Type_case.SPA.value
    z=None
    if not is_tomo:
        out = {k: list() for k in box_dict.particle_dictionary}
        for f in box_dict.particle_dictionary.keys():
            out[f] = fill(box_dict.get_sketches(f=f, n=None, is_3D=False, is_tracing=False))
    else:
        out = {tomoname: dict() for tomoname in box_dict.particle_dictionary}
        for tomoname,list_z in box_dict.particle_dictionary.items():
            for z in list_z:
                out[tomoname].update({z: fill(box_dict.get_sketches(f=tomoname, n=z, is_3D=False, is_tracing=False))})

    box_dict.fil_start_end_vis = starting_vis
    return out




"""
Script to add priors to particles attached to virons.

1.3.2022 - Thorsten Wagner
"""

from cryoloBM.bmtool import BMTool
from argparse import ArgumentParser
import numpy
import numpy as np
from eulerangles import matrix2euler
import pandas as pd
from typing import List
import argparse
from pyStarDB import sp_pystardb as star
import os

class SphericalPrior(BMTool):

    def create_parser(self, parentparser : ArgumentParser) -> ArgumentParser:

        helptxt = "Calculate priors of particles around a spherical model. It estimates 2 out of the 3 Euler angles a priori by calculating a vector from the centre of the sphere to each picked particle on the surface of the sphere."

        parser = parentparser.add_parser(
            self.get_command_name(),
            help=helptxt,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )

        req = parser.add_argument_group(
            "Arguments",
            helptxt,
        )

        req.add_argument("-s", "--star", required=True, help="Star file with picks for which the priors have to be calculated.")
        req.add_argument("-c", "--center", required=True, help=".coords file with center coordaintes of the spheres (e.g. virons).")
        req.add_argument("-o", "--out", required=True, help="output directory")

        return parser

    def get_command_name(self) -> str:
        '''
        :return: Command name
        '''
        return "sphericalprior"

    @staticmethod
    def read_coords(pth: str) -> np.array:
        return np.genfromtxt(pth)

    @staticmethod
    def rotation_matrix_from_vectors(vec1: np.array, vec2: np.array):
        """ Find the rotation matrix that aligns vec1 to vec2
        :param vec1: A 3d "source" vector
        :param vec2: A 3d "destination" vector
        :return mat: A transform matrix (3x3) which when applied to vec1, aligns it with vec2.
        """
        a, b = (vec1 / numpy.linalg.norm(vec1)).reshape(3), (vec2 / numpy.linalg.norm(vec2)).reshape(3)
        v = numpy.cross(a, b)
        if any(v):  # if not all zeros then
            c = numpy.dot(a, b)
            s = numpy.linalg.norm(v)
            kmat = numpy.array([[0, -v[2], v[1]], [v[2], 0, -v[0]], [-v[1], v[0], 0]])
            return numpy.eye(3) + kmat + kmat.dot(kmat) * ((1 - c) / (s ** 2))

        else:
            return numpy.eye(3)

    @staticmethod
    def vec_to_euler(center: np.array, coord: np.array):
        diff = coord - center
        ref = np.array([0, 0, 1])

        rot_mat = SphericalPrior.rotation_matrix_from_vectors(diff, ref)

        eulers = matrix2euler(rot_mat,
                              axes='zxz',
                              intrinsic=True,
                              right_handed_rotation=True)
        eulers[2] = eulers[2] + 90
        return eulers

    @staticmethod
    def group_by_center(centers: np.array, picks: np.array, max_distance: float) -> List[int]:
        """
        returns list of center index. -1 for invalid picks.
        """
        groups = []
        for row in picks:

            diff = centers - row
            diff = diff * diff
            diff = np.sum(diff,axis=1)
            diff = np.sqrt(diff)
            if np.min(diff) < max_distance:
                group_index = np.argmin(diff)
                groups.append(group_index)
            else:
                groups.append(-1)

        return groups

    @staticmethod
    def fill_priors(picks: pd.DataFrame, centers: np.array, groups: List[int]):
        tilts = []
        psis = []
        for row_index, row in picks.iterrows():
            coord = row[['_rlnCoordinateX', '_rlnCoordinateY', '_rlnCoordinateZ']].to_numpy().astype(float)
            center = centers[groups[row_index], :].astype(float)
            priors = SphericalPrior.vec_to_euler(center, coord)
            tilts.append(priors[1])
            psis.append(priors[2])
        if '_rlnAngleRot' not in picks:
            picks['_rlnAngleRot'] = 0
        picks['_rlnAngleTilt'] = tilts
        picks['_rlnAnglePsi'] = psis



    def run(self, args):
        pth_star = args.star
        pth_center = args.center
        pth_out = args.out

        coords = SphericalPrior.read_coords(pth_center)
        sfile = star.StarFile(pth_star)
        picks = sfile['']
        picks_rln = picks[['_rlnCoordinateX', '_rlnCoordinateY', '_rlnCoordinateZ']].to_numpy()
        coords = coords.astype(dtype=np.float)
        picks_rln = picks_rln.astype(dtype=np.float)
        groups= SphericalPrior.group_by_center(centers=coords, picks=picks_rln, max_distance=np.inf)

        SphericalPrior.fill_priors(picks=picks, centers=coords, groups=groups)

        os.makedirs(pth_out)
        new_sfile = star.StarFile(os.path.join(pth_out,os.path.basename(pth_star)))
        new_sfile.update('', picks, loop=True)
        new_sfile.write_star_file(overwrite=True)





from argparse import ArgumentParser
import argparse
from cryoloBM.bmtool import BMTool
from cryolo.utils import Filament, BoundBox
from cryolo.CoordsIO import read_cbox_boxfile, write_cbox_file
from dataclasses import dataclass
from typing import Tuple, List, Dict
import numpy as np
import os
import glob
import tqdm

@dataclass
class RotationMeta:
    filename_unrotated : str
    filename_rotated : str
    rotation : Tuple[float, float, float] # x,y,z
    dimension_orig : Tuple[int, int, int] # x,y,z
    dimension_rotated : Tuple[int, int, int] # x,y,z


class RotateCBOXCoords(BMTool):



    def run(self, args):
        cbox_path = args.input
        meta_path = args.meta
        output_path= args.out
        zfloat = args.zfloat
        extr_boxsize = args.boxsize

        cbox_files = self.get_cbox_files_paths(cbox_path)
        rotation_data = self.get_rotation_data(meta_path)
        matched_pairs = self.match_cbox_rotation_data(cbox_files=cbox_files, rotation_meta=rotation_data)
        for cbox_path, rotation_meta in tqdm.tqdm(matched_pairs, desc="Rotate"):
            cbox_dat = read_cbox_boxfile(cbox_path)
            is_filament = isinstance(cbox_dat[0], Filament)
            if is_filament:
                # filament path
                rotated_dat= self.rotate_filaments(filaments=cbox_dat, meta=rotation_meta,extraction_box_size=extr_boxsize)



            else:
                # bounding box path
                rotated_dat = self.rotate_boundingbox(boxes=cbox_dat, meta_path=rotation_meta, extraction_box_size=extr_boxsize)
                if not zfloat:
                    for box in rotated_dat:
                        box.z = int(round(box.z))

            filename = os.path.splitext(os.path.basename(rotation_meta.filename_unrotated))[0]

            out_cbox = os.path.join(output_path,"CBOX/")
            os.makedirs(out_cbox, exist_ok=True)
            out_file_pth = os.path.join(out_cbox, filename+".cbox")
            write_cbox_file(path=out_file_pth, coordinates=rotated_dat)


    @staticmethod
    def match_cbox_rotation_data(cbox_files: List[str], rotation_meta: List[RotationMeta]):
        pairs=[]
        for meta in rotation_meta:
            for pth in cbox_files:
                filename = os.path.splitext(os.path.basename(pth))[0]
                if filename in meta.filename_rotated:
                    pairs.append((pth,meta))
        return pairs

    @staticmethod
    def remove_out_of_volume_coords(coords: np.array, vol_dim : np.array, box_size = 0) -> np.array:
        mask = np.array([True]*coords.shape[0])
        for i in range(len(vol_dim)):
            mask_dim = np.logical_and(coords[:,i]>=box_size,coords[:,i] < (vol_dim[i] - box_size/2))
            mask = np.logical_and(mask,mask_dim)
        return coords[mask]

    def create_parser(self, parentparser: argparse.ArgumentParser) -> ArgumentParser:
        parser_cbox_rotate = parentparser.add_parser(
            self.get_command_name(),
            help="Given coordinates of from a rotated tomogram it rotates them back to a unrotated tomogram.",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )

        req = parser_cbox_rotate.add_argument_group(
            "Arguments",
            "Add filament prior information to star",
        )

        req.add_argument(
            "-i",
            "--input",
            required=True,
            help="Path to folder with CBOX files that are to rotate ",
        )

        req.add_argument(
            "-m",
            "--meta",
            required=True,
            help="Path to meta file",
        )

        req.add_argument(
            "-o",
            "--out",
            required=True,
            help="Path to output dir.",
        )

        req.add_argument(
            "--zfloat",
            default=False,
            action="store_true",
            help="If set, z is NOT converted to an integer"
        )

        req.add_argument(
            "--boxsize",
            default=-1,
            type=int,
            help="Extraction box size that should be taken into account when checking if rotated boxes are within the volume. If not provided, box size of CBOX files is used"
        )

        return parser_cbox_rotate


    def get_command_name(self) -> str:
        return "rotate"

    @staticmethod
    def get_rotation_data(meta: str) -> List[RotationMeta]:
        import pandas as pd
        dat = pd.read_csv(meta, header=None)
        meta_data=[]
        for _, row in dat.iterrows():
            meta = RotationMeta(
                filename_unrotated=row[0],
                filename_rotated=row[1],
                rotation=(float(np.radians(row[2])),
                          float(np.radians(row[3])),
                          float(np.radians(row[4]))),
                dimension_orig=(row[5],row[6],row[7]),
                dimension_rotated=(row[8],row[9],row[10]),
            )
            meta_data.append(meta)
        return meta_data

    @staticmethod
    def get_cbox_files_paths(pth: str):
        files = glob.glob(os.path.join(pth,"*.cbox"))
        return files

    def rotate_boundingbox(self, boxes : List[BoundBox], meta : RotationMeta, extraction_box_size = -1) -> List[BoundBox]:
        '''
        Rotates the bounding boxes according rotation meta
        '''

        # create numpy array with x,y,z, apply necessary conversions
        coords = np.zeros(shape=(len(boxes),3))
        avg_box_size = 0
        for i, b in enumerate(boxes):
            coords[i,0] = b.x + b.w/2
            coords[i,1] = b.y + b.h/2
            coords[i,2] = b.z
            avg_box_size += b.w
        avg_box_size = avg_box_size/len(boxes)
        if extraction_box_size != -1:
            avg_box_size = extraction_box_size



        rotated_boxes = []

        points = self.rotate_points(coords_raw=coords, meta=meta)

        points = RotateCBOXCoords.remove_out_of_volume_coords(points, meta.dimension_rotated, box_size=avg_box_size)


        for row_index, row in enumerate(points):
            box = BoundBox(x=row[0]-boxes[row_index].w/2,
                           y=row[1]-boxes[row_index].h/2,
                           z=row[2],
                           w=boxes[row_index].w,
                           h=boxes[row_index].h,
                           c=boxes[row_index].c,
                           depth=1)
            rotated_boxes.append(box)

        return rotated_boxes

    def rotate_filaments(self, filaments : List[Filament],
                         meta : RotationMeta,
                         extraction_box_size : int = 0
                         ) -> List[Filament]:
        '''
        Rotates the filament positions.
        '''

        rotated_filaments: List[Filament] = []

        for filament in filaments:
            boxes = filament.boxes
            rotated_boxes = [self.rotate_boundingbox(boxes=boxes,meta=meta,extraction_box_size=extraction_box_size)]
            for boxes in rotated_boxes:
                f = Filament(boxes=boxes)
                rotated_filaments.append(f)

        return rotated_filaments


    def rotate_points(self, coords_raw, meta : RotationMeta):

        # calc from meta
        thickness_rot = meta.dimension_rotated[2]
        thickness_ori= meta.dimension_orig[2]
        rot_angles_xyz = meta.rotation
        tomogram_dimensions_original = (meta.dimension_orig[0],meta.dimension_orig[1])
        # Translation to shift the origin to the center
        '''
        coords_columnsplit = np.hsplit(coords_raw, 4)
        coords = np.column_stack(
            (coords_columnsplit[0], coords_columnsplit[1], coords_columnsplit[2])
        )
        '''
        #print(coords_raw)
        coords = coords_raw
        recoords_matrix = np.array([-meta.dimension_rotated[0] / 2, -meta.dimension_rotated[1] / 2,
                                    -float(thickness_rot / 2)])
        coords_shifted = np.add(coords, recoords_matrix)

        # Invert coords array for rotation
        coords_inv = coords_shifted.swapaxes(0, 1)

        # Calculate rotation matrix
        rotx, roty, rotz = rot_angles_xyz
        R_z = np.array(
            [[np.cos(rotz), -np.sin(rotz), 0], [np.sin(rotz), np.cos(rotz), 0], [0, 0, 1]]
        )
        R_y = np.array(
            [[np.cos(roty), 0, np.sin(roty)], [0, 1, 0], [-np.sin(roty), 0, np.cos(roty)]]
        )
        R_x = np.array(
            [[1, 0, 0], [0, np.cos(rotx), -np.sin(rotx)], [0, np.sin(rotx), np.cos(rotx)]]
        )
        R = np.dot(R_x, np.dot(R_y, R_z))
        R_inv = np.linalg.inv(R)

        # Rotate the points
        new_coords_inv = np.dot(R_inv, coords_inv)

        # Revert coords and shift back the origin to the corner
        new_coords = new_coords_inv.swapaxes(0, 1)

        recoords_matrix_inv = np.array(
            [tomogram_dimensions_original[0] / 2, tomogram_dimensions_original[1] / 2,
             float(thickness_ori / 2)]
        )  # Half dimension orignal tomogram
        new_coords_recoords = np.add(new_coords, recoords_matrix_inv)
       # final = np.column_stack((new_coords_recoords))


        return new_coords_recoords


"""
The following functions manage the changes due to the modifications of the values of the table tracing

Which table?
"""

from cryoloBM import constants,  helper, visualization_3D_Window
from cryolo import grouping3d,  utils

from cryoloBM import boxmanager_controller, MySketch # for type hinting purpose
from typing import List, Union

def check_and_run_preview(controller:boxmanager_controller, only_reload:bool)->None:
    """
    :param controller: Boxmanager_controller obj
    :param only_reload: If True it plot the result on the current image
    """
    if controller.view.preview_is_on and not controller.model.has_filament:
        preview_run(controller=controller,only_reload=only_reload)

def searchRange_changed(controller:boxmanager_controller)->None:
    """
    :param controller: Boxmanager_controller obj
    """
    controller.model.trace_all = False
    purefilename = controller.get_current_filename(with_extension=False)
    value = int(controller.view.search_range_slider.value())
    try:
        if 0 <= value < controller.model.smallest_image_dim[purefilename] - 1:
            controller.model.search_range = value
        else:
            controller.view.err_message(msg= f"the value has to be in [0 - {controller.model.smallest_image_dim[purefilename] - 1}]")
        if not controller.view.is_updating_params and controller.model.type_case == constants.Type_case.TOMO_FOLDER.value and not controller.view.is_slider_pressed:
            update_all =controller.view.apply_to_all_the_tomo_question( name_param = "Search range") == controller.view.clicked_yes()
            if update_all:
                controller.model.trace_all = True
                for p in controller.model.params.values():
                    p.search_range = controller.model.search_range
            else:
                controller.model.params[purefilename].search_range = controller.model.search_range
    except ValueError:
        return

    controller.view.line_setText(line=controller.view.search_range_line,value=str(controller.model.search_range))
    check_and_run_preview(controller=controller,only_reload=True)

def searchRange_label_changed(controller:boxmanager_controller)->None:
    """
    :param controller: Boxmanager_controller obj
    """
    try:
        purefilename = controller.get_current_filename(with_extension=False)
        new_value = int(controller.view.search_range_line.text())
        if new_value > controller.model.smallest_image_dim[purefilename] or new_value < 0:
            return
    except ValueError:
        return
    controller.model.search_range = new_value
    controller.view.search_range_slider.setValue(new_value)
    check_and_run_preview(controller=controller, only_reload=True)

def memory_changed(controller:boxmanager_controller)->None:
    """
    :param controller: Boxmanager_controller obj
    """
    controller.model.trace_all = False
    purefilename = controller.get_current_filename(with_extension=False)
    value = int(controller.view.memory_slider.value())
    try:
        if 0 <= value < controller.model.tot_frames[purefilename] - 1:
            controller.model.memory = value
        else:
            controller.view.err_message(msg=f"the value has to be in [0 - {controller.model.tot_frames[purefilename] - 1}]")

        if not controller.view.is_updating_params and controller.model.type_case == constants.Type_case.TOMO_FOLDER.value and not controller.view.is_slider_pressed:
            update_all =controller.view.apply_to_all_the_tomo_question( name_param = "Memory") == controller.view.clicked_yes()
            if update_all:
                controller.model.trace_all = True
                for p in controller.model.params.values():
                    p.memory = controller.model.memory
            else:
                controller.model.params[purefilename].memory = controller.model.memory
    except ValueError:
        return

    controller.view.line_setText(line=controller.view.memory_line, value=str(controller.model.memory))
    check_and_run_preview(controller=controller,only_reload=True)

def memory_label_changed(controller:boxmanager_controller)->None:
    """
    :param controller: Boxmanager_controller obj
    """
    try:
        purefilename = controller.get_current_filename(with_extension=False)
        new_value = int(controller.view.memory_line.text())
        if new_value > controller.model.smallest_image_dim[purefilename] or new_value < 0:
            return
    except ValueError:
        return
    controller.model.memory = new_value
    controller.view .memory_slider.setValue(new_value)
    check_and_run_preview(controller=controller,only_reload=True)

def min_length_changed(controller:boxmanager_controller)->None:
    """
    :param controller: Boxmanager_controller obj
    """
    purefilename = controller.get_current_filename(with_extension=False)
    controller.model.trace_all = False
    value = int(controller.view.min_length_slider.value())
    try:
        if 0 <= value < controller.model.tot_frames[purefilename] - 1:
            controller.model.min_length = value
        else:
            controller.view.err_message(msg=f"the value has to be in [0 - {controller.model.tot_frames[purefilename] - 1}")

        if not controller.view.is_updating_params and controller.model.type_case == constants.Type_case.TOMO_FOLDER.value and not controller.view.is_slider_pressed:
            update_all = controller.view.apply_to_all_the_tomo_question(name_param="Minimum length") == controller.view.clicked_yes()
            if update_all:
                controller.model.trace_all = True
                for p in controller.model.params.values():
                    p.min_length = controller.model.min_length
            else:
                controller.model.params[purefilename].min_length = controller.model.min_length
    except ValueError:
        return

    controller.view.line_setText(line=controller.view.min_length_line, value=str(controller.model.min_length))
    check_and_run_preview(controller=controller,only_reload=True)

def min_length_label_changed(controller:boxmanager_controller)->None:
    """
    :param controller: Boxmanager_controller obj
    """
    try:
        purefilename = controller.get_current_filename(with_extension=False)
        new_value = int(controller.view.min_length_line.text())
        if new_value > controller.model.smallest_image_dim[purefilename] or new_value < 0:
            return
    except ValueError:
        return
    controller.model.min_length = new_value
    controller.view.min_length_slider.setValue(new_value)
    check_and_run_preview(controller=controller, only_reload=True)

def min_edge_weight_changed(controller:boxmanager_controller)->None:
    """
    :param controller: Boxmanager_controller obj
    """
    purefilename = controller.get_current_filename(with_extension=False)
    controller.model.trace_all = False
    try:
        controller.model.min_edge_weight = float(controller.view.min_edge_weight_slider.value()) / 100

        if not controller.view.is_updating_params and controller.model.type_case == constants.Type_case.TOMO_FOLDER.value and not controller.view.is_slider_pressed:
            update_all = controller.view.apply_to_all_the_tomo_question(name_param="Minimum edge weigh") == controller.view.clicked_yes()
            if update_all:
                controller.model.trace_all = True
                for p in controller.model.params.values():
                    p.min_edge_weight = controller.model.min_edge_weight
            else:
                controller.model.params[purefilename].min_edge_weight = controller.model.min_edge_weight
    except ValueError:
        return

    controller.view.line_setText(line=controller.view.min_length_line, value=str(controller.model.min_length))
    check_and_run_preview(controller=controller,only_reload=True)

def min_edge_weight_label_changed(controller:boxmanager_controller)->None:
    """
    :param controller: Boxmanager_controller obj
    """
    try:
        old_value = controller.model.min_edge_weight
        new_value = float(controller.view.min_edge_weight_line.text())
        if new_value > 1.0 or new_value < 0:
            controller.view.line_setText(line=controller.view.min_edge_weight_line, value=str(old_value))
            return
    except ValueError:
        return
    controller.model.min_edge_weight = new_value
    controller.view.min_edge_weight_slider.setValue(new_value * 100)

            #this
    #if self.preview_is_on:
    #    self.preview_win.set_list_patches(self.get_patches())
    #    self.preview_win.display_img(self.current_image_path, None, self.index_3D_tomo)

            # or this???
    check_and_run_preview(controller=controller, only_reload=True)

def win_size_changed(controller:boxmanager_controller)->None:
    """
    :param controller: Boxmanager_controller obj
    """
    controller.model.trace_all = False
    purefilename = controller.get_current_filename(with_extension=False)
    value = int(controller.view.win_size_slider.value())
    try:
        if value >= 0:
            controller.model.win_size = value
        else:
            controller.view.err_message(msg=f"the value has to be > 0")

        if not controller.view.is_updating_params and controller.model.type_case == constants.Type_case.TOMO_FOLDER.value and not controller.view.is_slider_pressed:
            update_all =controller.view.apply_to_all_the_tomo_question( name_param = "Window size") == controller.view.clicked_yes()
            if update_all:
                controller.model.trace_all = True
                for p in controller.model.params.values():
                    p.win_size = controller.model.win_size
            else:
                controller.model.params[purefilename].win_size = controller.model.win_size
    except ValueError:
        return

    controller.view.line_setText(line=controller.view.win_size_line, value=str(controller.model.win_size))
    check_and_run_preview(controller=controller,only_reload=True)

def win_size_label_changed(controller:boxmanager_controller)->None:
    """
    :param controller: Boxmanager_controller obj
    """
    try:
        new_value = float(controller.view.win_size_line.text())
        if new_value < 0:
            return
    except ValueError:
        return
    controller.model.win_size = new_value
    controller.view.win_size_slider.setValue(new_value)
    check_and_run_preview(controller=controller, only_reload=True)


def preview(controller:boxmanager_controller)->None:
    """
    :param controller: Boxmanager_controller obj
    """
    if controller.view.is_updating_params:
        return
    if controller.model.current_tomoimage_mmap is None:
        controller.view.err_message(msg="Please open a tomo image or a folder of tomo images first")
        return

    controller.view.preview_is_on = controller.view.preview_checkbox.isChecked()

    if controller.view.preview_is_on:
        preview_run(controller=controller,only_reload = False)

def trace(controller:boxmanager_controller)->None:
    """
    This button traces the particles and show them in the 'preview' 3D windows.
    The button is visible only if 'preview' checkbox is not checked
    :param controller: Boxmanager_controller obj
    """
    if controller.model.current_tomoimage_mmap is None:
        controller.view.err_message(msg="Please open a tomo image or a folder of tomo images first")
        return

    #check the preview checkbox if it is not checked
    if not controller.view.preview_is_on:
        controller.view.is_updating_params = True
        controller.view.check_preview_checkbox(check=True)
        controller.view.is_updating_params = False
        controller.view.preview_is_on = True

    controller.model.trace_all = False
    if controller.model.type_case == constants.Type_case.TOMO_FOLDER.value:
        if controller.view.qtmessage( msg="Do you want to apply the trace to all tomographies?",has_cancel=False) == controller.view.clicked_yes():
            controller.model.trace_all = True
            for p in controller.model.params.values():
                p.memory = controller.model.memory
                p.search_range = controller.model.search_range
                p.min_length = controller.model.min_length
                p.win_size = controller.model.win_size
                p.min_edge_weight = controller.model.min_edge_weight
    preview_run(controller=controller,only_reload = False)

def trace_values(controller:boxmanager_controller, fname:Union[str,None])->List[utils.BoundBox]:
    """
    :param controller: Boxmanager_controller obj
    :param fname: in folder case it is the name of the tomo on which perform the trace, otherwise None
    """
    if fname:
        param = controller.model.params[fname]
    else:
        fname = controller.get_current_filename(with_extension=False)
        param =helper.Params(fname, controller.model.filter_freq,
                             controller.model.boxsize,
                             controller.model.upper_size_thresh,
                             controller.model.lower_size_thresh,
                             controller.model.current_conf_thresh,
                             controller.model.current_num_boxes_thresh,
                             controller.model.min_edge_weight,
                             controller.model.search_range,
                             controller.model.memory,
                             controller.model.min_length,
                             controller.model.boxsize)


    if controller.model.has_filament:
        filament_dict = dict()#todo self.filament_dict[fname] if self.is_folder_3D_tomo else self.filament_dict
        return grouping3d.do_tracing_filaments(filament_dict, param.search_range, param.memory, param.conf_thresh,
                                               param.min_edge_weight, param.win_size)
    else:
        d = grouping3d.do_tracing(controller.model.box_dictionary.get_particle_dictionary_as_bbox(fname) , param.search_range, param.memory, param.conf_thresh, param.min_length)
        if d:
            d = utils.non_maxima_suppress_fast_3d(d, 0.3, param.conf_thresh)
    return d

def display_preview(controller:boxmanager_controller, only_reload:bool)->None:
    """
    Displays the results of the tracing.
    It is used in the confidence thresholding too, where i have just to filter the data
        :param controller: Boxmanager_controller obj
    :param only_reload: If True it show the result on the current image
    """
    patches = get_patches(controller=controller)
    if not controller.view.preview_win:
        # I should not be here after changing a threshold option
        controller.view.preview_win = visualization_3D_Window.PreviewWindow(controller.view.font, patches)
        only_reload = False
    else:
        controller.view.preview_win.set_list_patches(patches)

    if only_reload:
        controller.view.preview_win.reload_img()
    else:
        img = controller.model.current_tomoimage_mmap[controller.model.index_tomo, :, :]
        controller.view.preview_win.display_img(controller.get_current_filename(with_extension=True), img, controller.model.index_tomo)

def preview_run(controller: boxmanager_controller, only_reload: bool = False)->None:
    """
    This function will collect the params, run the tracing and show the result
    :param controller: Boxmanager_controller obj
    :param only_reload: If True it show the result on the current image
    """
    purefilename = controller.get_current_filename(with_extension=False)

    # run trace and convert the results from bboxes to sketches and save them into box_dictionary.box_dict_traced
    if controller.model.trace_all:
        for f in [*controller.model.box_dictionary.box_dict_traced]:
            controller.model.box_dictionary.box_dict_traced[f] = helper.convert_list_bbox_to_sketch(boxes_list=trace_values(controller=controller, fname=f),out_dict=True,col='r',fil_id=None)
    else:
        res = trace_values(controller=controller, fname=None)
        controller.model.box_dictionary.box_dict_traced[purefilename] =helper.convert_list_bbox_to_sketch(boxes_list=res,out_dict=True,col='r',fil_id=None)


    if not controller.model.has_filament:
        controller.model.box_dictionary.create_traced_3D()
    else:
        pass
        """
        if self.is_folder_3D_tomo:
            for f in [*self.box_dict_traced]:
                self.box_dictionary_3D_view.update({f: dict()})
                self.convert_filamentList_to_box_dictionary(self.box_dictionary_3D_view, self.box_dict_traced[f], f)
        else:
            self.convert_filamentList_to_box_dictionary(self.box_dictionary_3D_view, d)
        """

    display_preview(controller=controller, only_reload=only_reload)


def get_patches(controller: boxmanager_controller) -> List[MySketch.MySketch]:
    """
    :param controller: Boxmanager_controller obj
    """
    filename = controller.get_current_filename(with_extension=False)
    use_circle = controller.model.visualization == constants.Visualization_cb.CIRCLE.value
    current_conf_thresh, upper_size_thresh, lower_size_thresh, current_num_boxes_thresh = controller.model.get_threshold_params(filename)
    visible_rects = [box.getSketch(circle =use_circle) for box in controller.model.box_dictionary.get_sketches(f=filename,n=controller.model.index_tomo,is_3D=True,  is_tracing = False) if
                     helper.check_if_should_be_visible(box,current_conf_thresh, upper_size_thresh, lower_size_thresh, current_num_boxes_thresh, is_filament = controller.model.has_filament)]
    return visible_rects


"""
This file run the cryoloBM project. Its structure is based on the model-view-controller design pattern for theoretical
info visit https://en.wikipedia.org/wiki/Model%E2%80%93view%E2%80%93controller
It contains also  a parser for collecting data from CLI and the 'start_boxmanager' function, which create the object for
 the model-view-controller design pattern
"""


from cryoloBM import helper,boxmanager_view,boxmanager_controller,boxmanager_model,helper_GUI
from sys import argv as sys_argv

def start_boxmanager(image_dir:str, box_dir:str, wildcard:str, is_tomo:bool)->None:
    app = helper_GUI.QtG.QApplication(sys_argv)
    view = boxmanager_view.Boxmanager_view(font=app.font())
    model = boxmanager_model.Boxmanager_model(image_dir=image_dir, box_dir=box_dir, wildcard=wildcard, is_tomo=is_tomo)
    c = boxmanager_controller.Boxmanager_controller(view=view, model=model, app=app)
    c.run()

def run()->None:
    # collect args from cli
    args = helper.create_parser().parse_args()

    start_boxmanager(args.image_dir, args.box_dir, args.wildcard, is_tomo=args.is_tomo)


if __name__ == "__main__":
    run()



"""
In this file the 'PreviewWindow' class is defined.
This class is used for showing the preview after the tracing
"""

from os import path
from cryoloBM import helper_GUI
import matplotlib.pyplot as plt

from cryoloBM import MySketch # for type hinting purpose
from numpy import ndarray # for type hinting purpose
from typing import List,Tuple

class ChildDlg(helper_GUI.QtG.QDialog):
    is_visible = True
    def closeEvent(self, event:helper_GUI.plt_qtbackend)->None:
        self.is_visible = False

class PreviewWindow(helper_GUI.QtG.QMainWindow):
    def __init__(self, font:helper_GUI.QFont, list_patches:Tuple[MySketch.MySketch] = tuple(), title:str = "Preview")->None:
        super(PreviewWindow, self).__init__(None)

        self.title = title
        self.current_image_path = None
        self.current_pure_filename = None
        self.list_patches = list_patches
        self.index = None

        # SETUP QT
        self.font = font
        self.fig = None
        self.ax = None
        self.im = None
        self.plot = ChildDlg(self)
        self.img = None



    def set_current_image_paths(self, img_path:str)->None:
        """
        Set the img paths
        """
        self.current_image_path = img_path
        self.current_pure_filename = path.splitext(path.basename(self.current_image_path))[0]

    def set_list_patches(self,list_patches:List[MySketch.MySketch])->None:
        """
        Set the list_patches
        """
        if self.ax:
            self.ax.patches=[]      #delete all the patches that represent the patches on the last displayed slice
        self.list_patches = list_patches

    def display_img(self, img_path:str, img:ndarray, index:int, reload_img:bool = False)->None:
        """
        Set the variable to show an image. I copied some code from 'box_manager._set_first_time_img' function
        :param img_path: path of 'img' used to set the 'current_image_path' variable
        :param img: 2D numpy img
        :param index: slice index
        :param reload_img: If True we do not have to set self.img again
        :return: none
        """
        self.set_is_visible(True)
        if reload_img is False:
            if img is not None:
                #change image case
                self.img = img
            else:
                #boxmanager.onclick case
                img=self.img

        self.set_current_image_paths(img_path)
        self.index = index

        # Display the image
        if self.im is None:
            self.fig, self.ax = plt.subplots(1)
            self.ax.xaxis.set_visible(False)
            self.ax.yaxis.set_visible(False)
            self.fig.tight_layout()
            self.im = self.ax.imshow(img, origin="lower", cmap="gray", interpolation="Hanning")
            # The following variables have to set only once otherwise even if you change img it will not be displayed
            self.fig.canvas.draw()
            self.plot.canvas = helper_GUI.plt_qtbackend.FigureCanvasQTAgg(self.fig)
            layout = helper_GUI.QtG.QVBoxLayout()
            layout.addWidget(self.plot.canvas)
            self.plot.setLayout(layout)
        else:
            self.im.set_data(img)

        for b in self.list_patches:
            self.ax.add_patch(b)        # it clip the box on the fig image and not on the window

        self.plot.canvas.draw()
        self.plot.setWindowTitle(self.title+": "+self.current_pure_filename + " index: " + str(index))

        # move the pop up window, it avoids to overlap the boxmanager window
        resolution = helper_GUI.QtG.QDesktopWidget().screenGeometry()
        self.move(
            (resolution.width() / 2) + (self.frameSize().width()),
            (resolution.height() / 2) + (self.frameSize().height()),
            )

        self.plot.show()


    def reload_img(self)->None:
        self.display_img(self.current_image_path,self.img,self.index, True)

    def is_visible(self)->bool:
        return self.plot and self.plot.is_visible

    def set_is_visible(self,is_visible)->None:
        if self.plot:
            self.plot.is_visible = is_visible


import numpy as np
from eulerangles import matrix2euler

def rotation_matrix_from_vectors(vec1: np.array, vec2: np.array):
    """ Find the rotation matrix that aligns vec1 to vec2
    :param vec1: A 3d "source" vector
    :param vec2: A 3d "destination" vector
    :return mat: A transform matrix (3x3) which when applied to vec1, aligns it with vec2.
    """
    a, b = (vec1 / np.linalg.norm(vec1)).reshape(3), (vec2 / np.linalg.norm(vec2)).reshape(3)
    v = np.cross(a, b)
    if any(v):  # if not all zeros then
        c = np.dot(a, b)
        s = np.linalg.norm(v)
        kmat = np.array([[0, -v[2], v[1]], [v[2], 0, -v[0]], [-v[1], v[0], 0]])
        return np.eye(3) + kmat + kmat.dot(kmat) * ((1 - c) / (s ** 2))

    else:
        return np.eye(3)

def _main_():
    a = np.array([0, 1, 0])
    b = np.array([0, 0, 1])
    rot_mat = rotation_matrix_from_vectors(a, b)
    print("First rotation matrix:")
    print(rot_mat)

    eulers = matrix2euler(rot_mat,
                          axes='zyz',
                          intrinsic=True,
                          right_handed_rotation=True)
    print("matrix2euler output for first rotation matrix:", eulers)

    a = np.array([0.0000001, 1, 0])
    b = np.array([0, 0, 1])
    rot_mat = rotation_matrix_from_vectors(a, b)
    print("Second rotation matrix")
    print(rot_mat)

    eulers = matrix2euler(rot_mat,
                          axes='zyz',
                          intrinsic=True,
                          right_handed_rotation=True)
    print("matrix2euler output for second rotation matrix:",eulers)

if __name__ == "__main__":
    _main_()
